from datetime import datetime, timedelta
import pathlib
import re

import numpy as np
import pandas as pd
import h5py

import asilib
import asilib.utils as utils
import asilib.download as download


image_base_url = 'https://ergsc.isee.nagoya-u.ac.jp/psa-pwing/pub/raw/lamp/geoff/'
skymap_base_url = 'https://ergsc.isee.nagoya-u.ac.jp/psa-pwing/pub/raw/lamp/geoff/'
local_base_dir = asilib.config['ASI_DATA_DIR'] / 'lamp_phantom'


def lamp_phantom(
    location_code, time=None, time_range=None, redownload=False, missing_ok=True, alt=100
):
    """
    Create an Imager instance of LAMP's Phantom ASI.

    Parameters
    ----------
    location_code: str
        The ASI's location code (four letters).
    time: str or datetime.datetime
        A time to look for the ASI data at. Either time or time_range
        must be specified (not both or neither).
    time_range: list
        A length 2 list of string-formatted times or datetimes to bracket
        the ASI data time interval.
    redownload: bool
        If True, will download the data from the internet, regardless of
        wether or not the data exists locally (useful if the data becomes
        corrupted).
    missing_ok: bool
        Wether to allow missing data files inside time_range (after searching
        for them locally and online).
    alt: int
        The mapping altitude.

    Example
    -------

    Returns
    -------
    :py:meth:`~asilib.imager.Imager`
        A THEMIS ASI instance with the time stamps, images, skymaps, and metadata.
    """
    if location_code.lower() == 'vee':
        meta = {
            'array': 'lamp_phantom',
            'location': 'VEE',
            'lat': 67.0135,
            'lon': -146.406,
            'alt': 0.2,
            'cadence': 0.01,
            'resolution': (640, 640),
        }
    else:
        raise NotImplementedError

    _skymap = load_skymap(location_code, alt, redownload)
    skymap = {
        'lat': _skymap['Lat'],
        'lon': _skymap['Lon'],
        'alt': alt,
        'el': _skymap['El'],
        'az': _skymap['Az'],
        'path': _skymap['path'],
    }

    file_paths = _get_files(location_code, time, time_range, redownload, missing_ok)

    start_times = len(file_paths) * [None]
    end_times = len(file_paths) * [None]
    for i, file_path in enumerate(file_paths):
        date_match = re.search(r'\d{4}', file_path.name)
        start_times[i] = datetime.strptime(f'20220305_{date_match.group()}', '%Y%m%d_%H%M')
        end_times[i] = start_times[i] + timedelta(minutes=1)

    file_info = {
        'path': file_paths,
        'start_time': start_times,
        'end_time': end_times,
        'loader': lamp_reader,
    }
    if time_range is not None:
        file_info['time_range'] = time_range
    else:
        file_info['time'] = time
    return asilib.Imager(file_info, meta, skymap)


def _get_files(location_code, time, time_range, redownload, missing_ok):
    """
    Get the LAMP Phantom image file paths.
    """
    if (time is None) and (time_range is None):
        raise ValueError('time or time_range must be specified.')
    elif (time is not None) and (time_range is not None):
        raise ValueError('both time and time_range can not be simultaneously specified.')

    local_dir = local_base_dir / 'images' / location_code.lower()

    if time is not None:
        time = utils.validate_time(time)
        filename = f'narrow_{time.strftime("%H%M")}.mat'
        file_paths = list(pathlib.Path(local_dir).rglob(filename))

        if (len(file_paths) == 0) or redownload:
            d = download.Downloader(image_base_url + f'{filename}')
            file_paths = d.download(local_dir, redownload=redownload, stream=True)
        return file_paths

    # Find multiple image files.
    if time_range is not None:
        time_range = utils.validate_time_range(time_range)
        file_times = utils.get_filename_times(time_range, dt='minutes')
        file_paths = []

        for file_time in file_times:
            filename = f'narrow_{file_time.strftime("%H%M")}.mat'
            _local_file_path = list(pathlib.Path(local_dir).rglob(filename))

            if (len(_local_file_path) == 0) or redownload:
                d = download.Downloader(image_base_url + f'{filename}')
                try:
                    file_paths.append(d.download(local_dir, redownload=redownload, stream=True))
                except (FileNotFoundError, AssertionError, ConnectionError) as err:
                    if missing_ok and (
                        ('does not contain any hyper references containing' in str(err))
                        or ('Only one href is allowed' in str(err))
                        or ('error response' in str(err))
                    ):
                        continue
                    raise
        return file_paths


def lamp_reader(path):
    """
    Loads in the Phantom images chunk_size at a time to avoid a memory
    overflow. The .mat format is version 7.3 which are technically
    in the hdf5 format.
    """
    chunk_size = 600
    # The filename only contains the start hour and minute, so we calculate
    # t0 assuming the 2022-03-05 launch date.
    date_match = re.search(r'\d{4}', path.name)
    t0 = datetime.strptime(f'20220305_{date_match.group()}', '%Y%m%d_%H%M')

    with h5py.File(path, 'r') as f:
        image_keys = [key for key in f.keys() if 'images' in key]
        assert len(image_keys) == 1, f'{len(image_keys)} image keys found in the file {path}.'
        image_key = image_keys[0]
        dt = 60 / f[image_key].shape[0]

        for pos in range(0, f[image_key].shape[0], chunk_size):
            times = np.array([t0 + timedelta(seconds=dt * i) for i in range(pos, pos + chunk_size)])
            # ::-1 switches column to row major
            images = np.transpose(f[image_key][pos : pos + chunk_size], axes=(0, 2, 1))
            # images = f[image_key][pos:pos+chunk_size]
            yield times, images


def find_skymap(location_code, alt, redownload=True):
    """
    Find the path to the skymap file.
    """
    local_dir = local_base_dir / 'skymaps' / location_code.lower()

    # Check if the AzEl skymap is already downloaded.
    local_azel_paths = list(
        pathlib.Path(local_dir).rglob('Average_fps_im_-17700_to_-17600_AzEl.txt')
    )

    if (len(local_azel_paths) == 0) or redownload:
        # Download the AzEl skymaps.
        d = download.Downloader(skymap_base_url + 'Average_fps_im_-17700_to_-17600_AzEl.txt')
        local_azel_paths = d.download(local_dir, redownload=redownload)

    if len(local_azel_paths) != 1:
        raise FileNotFoundError(
            f'Unable to find the "Average_fps_im_-17700_to_-17600_AzEl.txt" LAMP skymap.'
        )

    # Check if the Lat/Lon skymap is already downloaded.
    local_latlon_paths = list(
        pathlib.Path(local_dir).rglob(f'Average_fps_im_-17700_to_-17600_LatLong{alt:03}km.txt')
    )

    if (len(local_latlon_paths) == 0) or redownload:
        # Download the LatLon skymaps.
        d = download.Downloader(
            skymap_base_url + f'Average_fps_im_-17700_to_-17600_LatLong{alt:03}km.txt'
        )
        local_latlon_paths = d.download(local_dir, redownload=redownload)

    if len(local_latlon_paths) != 1:
        raise FileNotFoundError(
            f'Unable to find the "Average_fps_im_-17700_to_-17600_AzEl.txt" LAMP skymap.'
        )
    return local_azel_paths[0], local_latlon_paths[0]


def load_skymap(location_code, alt, redownload):
    """
    Load the skymap file and apply the transformations.
    """
    azel_path, latlon_path = find_skymap(location_code, alt, redownload)
    skymap = _load_skymap(azel_path, latlon_path)
    # skymap = _tranform_longitude_to_180(skymap)
    # skymap = _transform_azimuth_to_180(skymap)
    # skymap = _flip_skymap(skymap)
    return skymap


def _load_skymap(azel_path, latlon_path):
    """
    A helper function to load a THEMIS skymap and transform it.
    """
    # Load the skymap file and convert it to a dictionary.
    azel_txt_data = pd.read_csv(azel_path, skipinitialspace=True, skiprows=29, delimiter=' ')
    latlon_txt_data = pd.read_csv(latlon_path, skipinitialspace=True, skiprows=29, delimiter=' ')
    skymap_dict = {
        'Az': azel_txt_data['Az'].to_numpy().reshape(640, 640),
        'El': azel_txt_data['El'].to_numpy().reshape(640, 640),
        'Lat': latlon_txt_data['Lat'].to_numpy().reshape(640, 640),
        'Lon': latlon_txt_data['Long'].to_numpy().reshape(640, 640),
        'path': [azel_path, latlon_path],
    }
    return skymap_dict


def _flip_skymap(skymap):
    """
    IDL is a column-major language while Python is row-major. This function
    tranposes the 2- and 3-D arrays to make them compatable with the images
    that are saved in row-major.
    """
    for key in skymap:
        if hasattr(skymap[key], 'shape'):
            shape = skymap[key].shape
            if (len(shape) == 2) and (shape[0] == shape[1]):
                skymap[key] = skymap[key][:, :]  # For Az/El maps.
    return skymap


def _tranform_longitude_to_180(skymap):
    """
    Transform the glo array from (0 -> 360) to (-180 -> 180).
    """
    # Don't take the modulus of NaNs
    valid_val_idx = np.where(~np.isnan(skymap['glo']))
    skymap['glo'][valid_val_idx] = np.mod(skymap['glo'][valid_val_idx] + 180, 360) - 180
    return skymap


def _transform_azimuth_to_180(skymap):
    """
    Transform the azm array from (-180 -> 180) to (0 -> 360).
    """
    # Don't take the modulus of NaNs
    valid_val_idx = np.where(~np.isnan(skymap['azm']))
    skymap['azm'][valid_val_idx] = np.mod(skymap['azm'][valid_val_idx], 360)
    return skymap


if __name__ == '__main__':
    import matplotlib.pyplot as plt
    import matplotlib.gridspec

    import asilib.map
    from asilib.asi.lamp_phantom import lamp_phantom

    asi = lamp_phantom(
        'vee',
        time=datetime(2022, 3, 5, 11, 32),
        redownload=False,
    )

    fig = plt.figure(figsize=(8, 4))
    gs = matplotlib.gridspec.GridSpec(1, 2, fig)
    ax = fig.add_subplot(gs[0, 0])
    bx = asilib.map.create_map(
        lon_bounds=(asi.meta['lon'] - 8, asi.meta['lon'] + 8),
        lat_bounds=(asi.meta['lat'] - 4, asi.meta['lat'] + 4),
        fig_ax=(fig, gs[0, 1]),
    )
    ax.axis('off')
    asi.plot_fisheye(ax=ax)
    asi.plot_map(ax=bx)
    plt.show()



"""
Test the simple and cartopy geographic map creation functions.
"""

import pytest
import matplotlib.testing.decorators

try:
    import cartopy

    cartopy_imported = True
except ImportError as err:
    # You can also get a ModuleNotFoundError if cartopy is not installed
    # (as compared to failed to import), but it is a subclass of ImportError.
    cartopy_imported = False

##########################################
########## TEST SIMPLE MAP ###############
##########################################


@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_create_simple_map_no_subplot'],
    tol=10,
    remove_text=True,
    extensions=['png'],
)
def test_create_simple_map_no_subplot():
    """
    Create a geographic map on one subplot.
    """
    import asilib.map
    import matplotlib.pyplot as plt

    ax = asilib.map.create_simple_map(lon_bounds=[0, 38], lat_bounds=[50, 75])
    ax.set_title('Generated via asilib.map.create_map()')
    return


@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_create_simple_map_equal_subplots'],
    tol=10,
    remove_text=True,
    extensions=['png'],
)
def test_create_simple_map_equal_subplots():
    """
    Create a geographic map on one subplot and random data on the second. The subplots
    have have unequal sizes.
    """
    import asilib.map
    import numpy as np
    import matplotlib.pyplot as plt

    fig = plt.figure(figsize=(6, 10))
    bx = asilib.map.create_simple_map(lon_bounds=[0, 38], lat_bounds=[50, 75], fig_ax=(fig, 211))
    cx = fig.add_subplot(2, 1, 2)
    cx.plot(np.arange(10), np.arange(10))
    fig.suptitle('Two subplots with equal sizes')
    return


@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_create_simple_map_unequal_subplots'],
    tol=10,
    remove_text=True,
    extensions=['png'],
)
def test_create_simple_map_unequal_subplots():
    """
    Create a geographic map on one subplot and random data on the second. The subplots
    have have unequal sizes.
    """
    import asilib.map
    import numpy as np
    import matplotlib.pyplot as plt

    fig = plt.figure(figsize=(6, 10))
    dx = (3, 1, (1, 2))
    dx = asilib.map.create_simple_map(lon_bounds=[0, 38], lat_bounds=[50, 75], fig_ax=(fig, dx))
    ex = fig.add_subplot(3, 1, 3)
    ex.plot(np.arange(10), np.arange(10))
    fig.suptitle('Two subplots with unequal sizes')
    return


@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_create_simple_map_gridspec'],
    tol=10,
    remove_text=True,
    extensions=['png'],
)
def test_create_simple_map_gridspec():
    """
    Create a geographic map in a subplot created using gridspec.
    """
    import asilib.map
    import matplotlib.pyplot as plt
    import matplotlib.gridspec

    fig = plt.figure()
    gs = matplotlib.gridspec.GridSpec(1, 1, fig)
    dx = asilib.map.create_simple_map(lon_bounds=[0, 38], lat_bounds=[50, 75], fig_ax=(fig, gs))
    dx.set_title('Map made using gridspec')
    return


##########################################
########## TEST CARTOPY MAP ##############
##########################################


@pytest.mark.skipif(not cartopy_imported, reason='cartopy not installed')
@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_create_cartopy_map_no_subplot'],
    tol=10,
    remove_text=True,
    extensions=['png'],
)
def test_create_cartopy_map_no_subplot():
    """
    Create a geographic map on one subplot.
    """
    import asilib.map
    import matplotlib.pyplot as plt

    ax = asilib.map.create_cartopy_map(lon_bounds=[0, 38], lat_bounds=[50, 75])
    ax.set_title('Generated via asilib.map.create_cartopy_map()')
    return


@pytest.mark.skipif(not cartopy_imported, reason='cartopy not installed')
@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_create_cartopy_map_equal_subplots'],
    tol=10,
    remove_text=True,
    extensions=['png'],
)
def test_create_cartopy_map_equal_subplots():
    """
    Create a geographic map on one subplot and random data on the second. The subplots
    have have unequal sizes.
    """
    import asilib.map
    import numpy as np
    import matplotlib.pyplot as plt

    fig = plt.figure(figsize=(6, 10))
    bx = asilib.map.create_cartopy_map(lon_bounds=[0, 38], lat_bounds=[50, 75], fig_ax=(fig, 211))
    cx = fig.add_subplot(2, 1, 2)
    cx.plot(np.arange(10), np.arange(10))
    fig.suptitle('Two subplots with equal sizes')
    return


@pytest.mark.skipif(not cartopy_imported, reason='cartopy not installed')
@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_create_cartopy_map_unequal_subplots'],
    tol=10,
    remove_text=True,
    extensions=['png'],
)
def test_create_cartopy_map_unequal_subplots():
    """
    Create a geographic map on one subplot and random data on the second. The subplots
    have have unequal sizes.
    """
    import asilib.map
    import numpy as np
    import matplotlib.pyplot as plt

    fig = plt.figure(figsize=(6, 10))
    dx = (3, 1, (1, 2))
    dx = asilib.map.create_cartopy_map(lon_bounds=[0, 38], lat_bounds=[50, 75], fig_ax=(fig, dx))
    ex = fig.add_subplot(3, 1, 3)
    ex.plot(np.arange(10), np.arange(10))
    fig.suptitle('Two subplots with unequal sizes')
    return


@pytest.mark.skipif(not cartopy_imported, reason='cartopy not installed')
@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_create_cartopy_map_gridspec'],
    tol=10,
    remove_text=True,
    extensions=['png'],
)
def test_create_cartopy_map_gridspec():
    """
    Create a geographic map in a subplot created using gridspec.
    """
    import asilib.map
    import matplotlib.pyplot as plt
    import matplotlib.gridspec

    fig = plt.figure()
    gs = matplotlib.gridspec.GridSpec(1, 1, fig)
    dx = asilib.map.create_cartopy_map(lon_bounds=[0, 38], lat_bounds=[50, 75], fig_ax=(fig, gs))
    dx.set_title('Map made using gridspec')
    return



"""
The Imagers class handles multiple Imager objects and coordinates plotting and animating multiple
fisheye lens and mapped images. The mapped images are also called mosaics.
"""

from typing import Tuple, List, Union, Generator, Callable
from collections import namedtuple
import pathlib
from datetime import datetime, timedelta
from itertools import zip_longest
import shutil
import warnings

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors
import matplotlib.collections
import matplotlib.patches
import scipy.interpolate
from scipy.spatial import cKDTree
try:
    import IRBEM
    _irbem_imported = True
except ImportError:
    _irbem_imported = False

from asilib.imager import Imager, _haversine, Skymap_Cleaner
import asilib.map
import asilib.utils


class Imagers:
    """
    Plot and animate multiple :py:meth:`~asilib.imager.Imager` objects.

    .. warning::

        This class is in development and not all methods are implemented/matured.

    Parameters
    ----------
    imagers: Tuple
        The Imager objects to plot and animate. 
    sync_image_tol: float
        The allowable time tolerance, in units of a fraction of each imagers's cadence, 
        for images to be considered synchronized. Adjusting this kwarg is useful if 
        the imager has missing data and you need to animate a mosaic.
    """
    def __init__(self, imagers:Tuple[Imager], sync_image_tol:float=0.2) -> None:
        self.imagers = imagers
        # Wrap self.imagers in a tuple if the user passes in a single Imager object.
        if isinstance(self.imagers, Imager):
            self.imagers = (self.imagers, )
        self.sync_image_tol = sync_image_tol 

        if 'time_range' in self.imagers[0].file_info:
            assert np.all(['time_range' in img.file_info for img in self.imagers]), (
                'Not all imagers have a time_range defined.'
                )
            start_times = [img.file_info['time_range'][0] for img in self.imagers]
            end_times = [img.file_info['time_range'][1] for img in self.imagers]
            self.min_cadence = np.min([img.meta['cadence'] for img in self.imagers]).astype(float)

            if not all(start_time is None for start_time in start_times):
                self.start_time = min(start_times)
                self.end_time = max(end_times)
            self.n_times = int(np.ceil((self.end_time - self.start_time).total_seconds() / self.min_cadence))

        self.lon_bounds, self.lat_bounds = self.spatial_extent()
        return
    
    def spatial_extent(self) -> Tuple[Tuple[float, float], Tuple[float, float]]:
        """
        Calculate the longitude and latitude bounds of the mosaic.

        Returns
        -------
        Tuple[float, float]
            Longitude bounds.
        Tuple[float, float]
            Latitude bounds.
        """
        lon_bounds = [np.nan, np.nan]
        lat_bounds = [np.nan, np.nan]
        for imager in self.imagers:
            _imager_lon_bounds = np.nanmin(imager.skymap['lon']), np.nanmax(imager.skymap['lon'])
            _imager_lat_bounds = np.nanmin(imager.skymap['lat']), np.nanmax(imager.skymap['lat'])

            lon_bounds[0] = np.nanmin([lon_bounds[0], _imager_lon_bounds[0]])
            lon_bounds[1] = np.nanmax([lon_bounds[1], _imager_lon_bounds[1]])
            lat_bounds[0] = np.nanmin([lat_bounds[0], _imager_lat_bounds[0]])
            lat_bounds[1] = np.nanmax([lat_bounds[1], _imager_lat_bounds[1]])
        return tuple(lon_bounds), tuple(lat_bounds)
    
    def plot_fisheye(self, ax:Tuple[plt.Axes], **kwargs):
        """
        Plots one fisheye image in each subplot, oriented with North on the top, and East on the 
        left of the image.

        Parameters
        ----------
        ax: Tuple[plt.Axes]
            Subplots corresponding to each fisheye lens image.
        kwargs: dict
            Keyword arguments directly passed into each :py:meth:`~asilib.Imager.plot_fisheye`
            method.

        Example
        -------
        >>> from datetime import datetime
        >>> 
        >>> import matplotlib.pyplot as plt
        >>> import asilib
        >>> import asilib.asi
        >>> 
        >>> time = datetime(2007, 3, 13, 5, 8, 45)
        >>> location_codes = ['FSIM', 'ATHA', 'TPAS', 'SNKQ']
        >>> 
        >>> fig, ax = plt.subplots(1, len(location_codes), figsize=(12, 3.5))
        >>> 
        >>> _imagers = []
        >>> 
        >>> for location_code in location_codes:
        >>>     _imagers.append(asilib.asi.themis(location_code, time=time))
        >>> 
        >>> for ax_i in ax:
        >>>     ax_i.axis('off')
        >>> 
        >>> asis = asilib.Imagers(_imagers)
        >>> asis.plot_fisheye(ax=ax)
        >>> 
        >>> plt.suptitle('Donovan et al. 2008 | First breakup of an auroral arc')
        >>> plt.tight_layout()
        >>> plt.show()
        """
        if not isinstance(ax, (list, tuple, np.ndarray)):
            ax = (ax,)
        assert len(ax) == len(self.imagers), 'Number of subplots must equal the number of imagers.'

        for ax_i, imager_i in zip(ax, self.imagers):
            imager_i.plot_fisheye(ax=ax_i, **kwargs)
        return
    
    def plot_map(self, overlap=False, **kwargs):
        """
        Projects multiple ASI images onto a map at an altitude that is defined in the skymap 
        calibration file.

        Parameters
        ----------
        overlap: bool
            If True, pixels that overlap between imager FOV's are overplotted such that only the 
            final imager's pixels are shown.
        kwargs: dict
            Keyword arguments directly passed into each :py:meth:`~asilib.Imager.plot_map`
            method.

        Example
        -------
        >>> from datetime import datetime
        >>> 
        >>> import matplotlib.pyplot as plt
        >>> 
        >>> import asilib
        >>> import asilib.map
        >>> import asilib.asi
        >>> 
        >>> time = datetime(2007, 3, 13, 5, 8, 45)
        >>> location_codes = ['FSIM', 'ATHA', 'TPAS', 'SNKQ']
        >>> map_alt = 110
        >>> min_elevation = 2
        >>> 
        >>> ax = asilib.map.create_map(lon_bounds=(-140, -60), lat_bounds=(40, 82))
        >>> 
        >>> _imagers = []
        >>> 
        >>> for location_code in location_codes:
        >>>     _imagers.append(asilib.asi.themis(location_code, time=time, alt=map_alt))
        >>> 
        >>> asis = asilib.Imagers(_imagers)
        >>> asis.plot_map(ax=ax, overlap=False, min_elevation=min_elevation)
        >>> 
        >>> ax.set_title('Donovan et al. 2008 | First breakup of an auroral arc')
        >>> plt.show()
        """
        if 'ax' not in kwargs:
            kwargs['ax'] = asilib.map.create_map()

        # If the kwarg is not specified, or is None, then it should just be None for all imagers.
        # Below will be True if the kwarg['color_bounds'] is set as None, or if the key is missing.
        if kwargs.get('color_bounds', None) is None:
            color_bounds = np.full(len(self.imagers), None)
        elif len(np.array(kwargs['color_bounds']).shape) == 1:  # Same color_bounds for all imagers.
            color_bounds = np.repeat(
                np.array(kwargs['color_bounds'])[:, np.newaxis], 
                len(self.imagers), 
                axis=-1
                ).T
        elif len(np.array(kwargs['color_bounds']).shape) == 2:
            assert np.array(kwargs['color_bounds']).shape == (len(self.imagers), 2), (
                "A unique color_bounds must be specified for each imager."
                )
            color_bounds = kwargs['color_bounds']
        else:
            raise ValueError(
                f'The shape of color_bounds must be (1, 2), or '
                f'({len(self.imagers)}, 2), not {np.array(kwargs["color_bounds"]).shape}'
                )
        # Must remove the color_bounds variable, if the kwargs dict contains it, 
        # so the call to imager.plot_map() doesn't raise an error because of
        # duplicate keyword arguments. 
        kwargs.pop('color_bounds', None)

        self._skymaps = {
            _imager.meta['location']:{
                'lon':_imager.skymap['lon'].copy(), 
                'lat':_imager.skymap['lat'].copy(),
                'el':_imager.skymap['el'].copy()
                } for _imager in self.imagers
            }
        # If overlap=False, each skymap first has nans in the imager overlapping region,
        # and then the skymap cleaner nans everything below min_elevation, which is then
        # immediately followed by the reassignment of all nans to the nearest valid value.
        if not overlap:
            self._skymaps = self.nan_overlap_pixels(self._skymaps, shrink_distance_to_self=0.96)

        for imager, _color_bounds in zip(self.imagers, color_bounds):
            _skymap_cleaner = Skymap_Cleaner(
                self._skymaps[imager.meta['location']]['lon'], 
                self._skymaps[imager.meta['location']]['lat'], 
                self._skymaps[imager.meta['location']]['el'], 
            )
            _skymap_cleaner.mask_elevation(kwargs.get('min_elevation', 5))
            _cleaned_lon_grid, _cleaned_lat_grid = _skymap_cleaner.remove_nans()
            imager.plot_map(
                **kwargs, 
                color_bounds=_color_bounds, 
                lon_grid=_cleaned_lon_grid, 
                lat_grid=_cleaned_lat_grid
                )
        return
    
    # def animate_fisheye(self):
    #     raise NotImplementedError
    
    # def animate_fisheye_gen(self):
    #     raise NotImplementedError
    
    def animate_map(self, **kwargs):
        """
        Animate an ASI mosaic. It is a wrapper for the 
        :py:meth:`~asilib.Imagers.animate_map_gen` method.

        See :py:meth:`~asilib.Imagers.animate_map_gen` documentation for the complete 
        list of kwargs.

        Example
        -------
        .. code-block:: python

            >>> import asilib
            >>> import asilib.asi
            >>> 
            >>> time_range = ('2021-11-04T06:55', '2021-11-04T07:05')
            >>> asis = asilib.Imagers(
            >>>     [asilib.asi.trex_rgb(location_code, time_range=time_range) 
            >>>     for location_code in ['LUCK', 'PINA', 'GILL', 'RABB']]
            >>>     )
            >>> asis.animate_map(lon_bounds=(-115, -85), lat_bounds=(43, 63), overwrite=True)
        """
        for _ in self.animate_map_gen(**kwargs):
            pass
        return
    
    def animate_map_gen(
        self,
        overlap=False,
        timestamp: bool = True,
        lon_bounds: tuple = (-160, -50),
        lat_bounds: tuple = (40, 82),
        ax: Union[plt.Axes, tuple] = None,
        coast_color: str = 'k',
        land_color: str = 'g',
        ocean_color: str = 'w',
        color_map: str = None,
        color_bounds: List[float] = None,
        color_norm: str = None,
        min_elevation: float = 10,
        pcolormesh_kwargs: dict = {},
        asi_label: bool = True,
        movie_container: str = 'mp4',
        animation_save_dir: Union[pathlib.Path, str]=None,
        ffmpeg_params={},
        overwrite: bool = False,
        ) -> Generator[
            Tuple[datetime, list[datetime], list[np.ndarray], plt.Axes], None, None
        ]:
        """
        Animate an ASI mosaic.

        Parameters
        ----------
        overlap: bool
            If True, pixels that overlap between imager FOV's are overplotted such that only the 
            final imager's pixels are shown.
        timestamp: bool
            If True, the time stamp is annotated in the upper left corner of the map.
        lon_bounds: tuple
            The map's longitude bounds.
        lat_bounds: tuple
            The map's latitude bounds.
        ax: plt.Axes, tuple
            The subplot to put the map on. If cartopy is installed, ```ax``` must be
            a two element tuple specifying the ``plt.Figure`` object and subplot position
            passed directly as ``args`` into
            `fig.add_subplot() <https://matplotlib.org/stable/api/figure_api.html#matplotlib.figure.Figure.add_subplot>`_.
        coast_color: str
            The coast color. If None will not draw it.
        land_color: str
            The land color. If None will not draw it.
        ocean_color: str
            The ocean color. If None will not draw it.
        color_map: str
            The matplotlib colormap to use. See `matplotlib colormaps <https://matplotlib.org/stable/tutorials/colors/colormaps.html>`_
            for supported colormaps.
        color_bounds: List[float]
            The lower and upper values of the color scale. The default is: low=1st_quartile and
            high=min(3rd_quartile, 10*1st_quartile). This range works well for most cases.
        color_norm: str
            Set the 'lin' (linear) or 'log' (logarithmic) color normalization. If color_norm=None,
            the color normalization will be taken from the ASI array (if specified), and if not
            specified it will default to logarithmic.
        min_elevation: float
            Masks the pixels below min_elevation degrees.
        pcolormesh_kwargs: dict
            A dictionary of keyword arguments (kwargs) to pass directly into
            plt.pcolormesh.
        asi_label: bool
            Annotates the map with the ASI code in the center of the mapped image.
        movie_container: str
            The movie container: mp4 has better compression but avi was determined
            to be the official container for preserving digital video by the
            National Archives and Records Administration.
        ffmpeg_params: dict
            The additional/overwitten ffmpeg output parameters. The default parameters are:
            framerate=10, crf=25, vcodec=libx264, pix_fmt=yuv420p, preset=slower.
        overwrite: bool
            Overwrite the animation. If False, ffmpeg will prompt you to answer y/n if the
            animation already exists.

        Yields
        ------
        datetime.datetime
            The guide time used to keep the images synchronized.
        List[datetime.datetime]
            Nearest imager time stamps to the guide time. If the difference between
            the imager time and the guide time is greater than time_tol*imager_cadence, 
            or the imager is off, the imager is considered unsynchronized and the 
            returned time is datetime.min.
        List[np.ndarray]
            The images corresponding to the times returned above. If the that imager is
            unsynchronized, the corresponding image value is None.
        plt.Axes
            The subplot object.

        Example
        -------
        .. code-block:: python

            >>> # Animate a TREX-RGB mosaic and print the individual time stamps
            >>> # to confirm that the imagers are synchronized.
            >>> import asilib
            >>> import asilib.asi
            >>> 
            >>> time_range = ('2021-11-04T06:55', '2021-11-04T07:05')
            >>> asis = asilib.Imagers(
            >>>     [asilib.asi.trex_rgb(location_code, time_range=time_range) 
            >>>     for location_code in ['LUCK', 'PINA', 'GILL', 'RABB']]
            >>>     )
            >>> gen = asis.animate_map_gen(
            >>>     lon_bounds=(-115, -85), lat_bounds=(43, 63), overwrite=True
            >>>     )
            >>> for guide_time, asi_times, asi_images, ax in gen:
            >>>     if '_text_obj' in locals():
            >>>         _text_obj.remove()
            >>>     info_str = f'Guide: {guide_time: %Y:%m:%d %H:%M:%S}\\n'
            >>>     # The below for loop is possible because the imagers and 
            >>>     # asi_times can be indexed together.
            >>>     for _imager, _imager_time in zip(asis.imagers, asi_times):
            >>>         info_str += f'{_imager.meta["location"]}: {_imager_time: %Y:%m:%d %H:%M:%S}\\n'
            >>>     info_str = info_str[:-1]  # Remove the training newline
            >>> 
            >>>     _text_obj = ax.text(
            >>>         0.01, 0.99, info_str, va='top', transform=ax.transAxes, 
            >>>         bbox=dict(facecolor='grey', edgecolor='black'))
            
        """
        if ax is None:
            ax = asilib.map.create_map(
                lon_bounds=lon_bounds,
                lat_bounds=lat_bounds,
                coast_color=coast_color,
                land_color=land_color,
                ocean_color=ocean_color,
            )

        # Handle manual color_bounds for none, all, or each imager.
        if color_bounds is None:
            color_bounds = np.full(len(self.imagers), None)
        elif len(np.array(color_bounds).shape) == 1:  # Same color_bounds for all imagers.
            color_bounds = np.repeat(
                np.array(color_bounds)[:, np.newaxis], 
                len(self.imagers), 
                axis=-1
                ).T
        elif len(np.array(color_bounds).shape) == 2:
            assert np.array(color_bounds).shape == (len(self.imagers), 2), (
                "A unique color_bounds must be specified for each imager."
                )
        else:
            raise ValueError(
                f'The shape of color_bounds must be (1, 2), or '
                f'({len(self.imagers)}, 2), not {np.array(color_bounds).shape}'
                )

        # Create the animation directory inside asilib.config['ASI_DATA_DIR'] if it does
        # not exist.
        if animation_save_dir is None:
            _path = asilib.config['ASI_DATA_DIR']
        else:
            _path = animation_save_dir
        image_save_dir = pathlib.Path(
            _path,
            'animations',
            'images',
            f'{self.imagers[0].file_info["time_range"][0].strftime("%Y%m%d_%H%M%S")}_mosaic',
        )

        self.animation_name = (
            f'{self.imagers[0].file_info["time_range"][0].strftime("%Y%m%d_%H%M%S")}_'
            f'{self.imagers[0].file_info["time_range"][-1].strftime("%H%M%S")}_mosaic.'
            f'{movie_container}'
        )
        movie_save_path = image_save_dir.parents[1] / self.animation_name
        # If the image directory exists we need to first remove all of the images to avoid
        # animating images produced by different method calls.
        if image_save_dir.is_dir():
            shutil.rmtree(image_save_dir)
        image_save_dir.mkdir(parents=True)

        image_paths = []
        _progressbar = asilib.utils.progressbar(
            enumerate(self.__iter__()),
            iter_length=self.n_times,
            text=self.animation_name,
        )
        operating_asis = np.array([])

        for i, (_guide_time, _asi_times, _asi_images) in _progressbar:
            asi_labels = len(self.imagers)*[None]
            pcolormesh_objs = len(self.imagers)*[None]

            # Determine if an imager turned off/on and we need to recalculate overlapping skymaps.
            currently_on_asis = np.where(np.array(_asi_times) != datetime.min)[0].astype(int)
            if (currently_on_asis.shape != operating_asis.shape) or np.any(currently_on_asis != operating_asis):
                _skymaps = {}
                for j in currently_on_asis:
                    imager = self.imagers[j]
                    _skymaps[imager.meta['location']] = {
                        'lon':imager.skymap['lon'].copy(), 
                        'lat':imager.skymap['lat'].copy()
                    }

                if not overlap:
                    _skymaps = self.nan_overlap_pixels(_skymaps, idx=currently_on_asis)

                for j in currently_on_asis:
                    imager = self.imagers[j]
                    _skymap_cleaner = Skymap_Cleaner(
                        _skymaps[imager.meta['location']]['lon'], 
                        _skymaps[imager.meta['location']]['lat'], 
                        imager.skymap['el'],
                    )
                    _skymap_cleaner.mask_elevation(min_elevation=min_elevation)
                    _cleaned_lon_grid, _cleaned_lat_grid = _skymap_cleaner.remove_nans()
                    _skymaps[imager.meta['location']]['lon'] = _cleaned_lon_grid
                    _skymaps[imager.meta['location']]['lat'] = _cleaned_lat_grid
                operating_asis = currently_on_asis

            for j, (_asi_time, _asi_image) in enumerate(zip(_asi_times, _asi_images)):
                if _asi_time == datetime.min:
                    continue
                _color_map, _color_norm = self.imagers[j]._plot_params(
                    _asi_image, color_bounds[j], color_map, color_norm
                    )

                ax, pcolormesh_objs[j], asi_labels[j] = self.imagers[j]._plot_mapped_image(
                    ax, 
                    _asi_image, 
                    min_elevation, 
                    _color_map, 
                    _color_norm, 
                    asi_label, 
                    pcolormesh_kwargs, 
                    lon_grid=_skymaps[self.imagers[j].meta['location']]['lon'], 
                    lat_grid=_skymaps[self.imagers[j].meta['location']]['lat']
                )

            if timestamp:
                if '_text_obj' in locals():
                    _text_obj.remove()  # noqa: F821
                _text_obj = ax.text(
                    0.01, 
                    0.99, 
                    f'{_guide_time: %Y-%m-%d %H:%M:%S}', 
                    va='top', 
                    transform=ax.transAxes,
                    fontsize=12,
                    color='white',
                    bbox=dict(facecolor='grey', edgecolor='black', alpha=0.5)
                )

            # Give the user the control of the subplot, image object, and return the image time
            # so that they can manipulate the image to add, for example, the satellite track.
            yield _guide_time, _asi_times, _asi_images, ax

            # Save the plot before the next iteration.
            save_name = f'{str(i).zfill(6)}.png'
            plt.savefig(image_save_dir / save_name)
            image_paths.append(image_save_dir / save_name)

            # Clean up the objects that this method generated.
            for _asi_label in asi_labels:
                if _asi_label is not None:
                    _asi_label.remove()
            for pcolormesh_obj in pcolormesh_objs:
                if pcolormesh_obj is not None:
                    pcolormesh_obj.remove()
        
        self.imagers[0]._create_animation(image_paths, movie_save_path, ffmpeg_params, overwrite)
        return
    
    def map_eq(self, b_model: Callable='IGRF') -> Tuple[np.ndarray, np.ndarray]:
        """
        Map an auroral image to the magnetic equator using IGRF or a user-defined magnetic field model.

        Parameters
        ----------
        b_model: Callable
            A function with the (time, lla) arguments where time is a single time stamp, 
            and lla is one (lat, lon, alt) point. If ``b_model='IGRF'``, IRBEM's IGRF model is used
            in the default GSM coordinates.

        Returns
        -------
        numpy.ndarray:
            Mapped locations in the Geocentric Solar Magnetospheric (GSM) coordinates.
        numpy.ndarray:
            Auroral pixel intensities.
        """
        warnings.warn(
            'asilib.Imagers.map_eq() is an experimental method. Its functionality or interface '
            'stability is not guaranteed.'
            )
        if b_model == 'IGRF':
            b_model = self._igrf_model_wrapper
            if not _irbem_imported:
                raise ImportError(
                    'IRBEM is not installed. Please install it to use the IGRF model, '
                    'or supply a custom b_model.'
                    )
            
        lats_lons, intensities = self.get_points()
        equator_sm = np.zeros((lats_lons[:, 0].shape[0], 3))

        _progressbar = asilib.utils.progressbar(
            enumerate(zip(lats_lons[:, 0], lats_lons[:, 1])),
            iter_length=lats_lons[:, 0].shape[0],
            text='Mapping to magnetic equator'
            )

        for i, (lat, lon) in _progressbar:
            equator_sm[i, :] = b_model(self.imagers[0].file_info['time'], (lat, lon, self.imagers[0].skymap['alt']))
        return equator_sm, intensities
    
    def plot_map_eq(
            self, 
            ax:plt.Axes=None, 
            b_model:Callable="IGRF",
            max_valid_grid_distance:float=0.03,
            x_grid:np.ndarray=None, 
            y_grid:np.ndarray=None, 
            color_bounds: List[float] = None,
            color_norm: str = None,
            color_map: str = None,
            pcolormesh_kwargs:dict={},
            ) -> Tuple[plt.Axes, matplotlib.collections.QuadMesh]:
        """
        Plot the equatorial projection of an auroral image, mapped using IGRF or a user-defined 
        magnetic field model.

        Parameters
        ----------
        ax: plt.Axes
            The subplot to put the map on. If None, one will be created.
        b_model: Callable
            A function with the (time, lla) arguments where time is a single time stamp, 
            and lla is one (lat, lon, alt) point. If ``b_model='IGRF'``, IRBEM's IGRF model is used
            in the default GSM coordinates.
        max_valid_grid_distance: float
            The maximum distance from the mapped point to the grid point. If the distance is greater
            than this value, the mapped grid point will be masked as NaN. The distance is in units of Re.
        x_grid, x_grid: np.ndarray
            The x and y grid points to plot the mapped image. If None, a grid of 1000x1001 points spanning 
            -12 < x < 1.1 and -5 < y < 5 is generated.
        color_bounds: List[float]
            The ASI image color bounds.
        color_map: str
            The matplotlib colormap to use. See `matplotlib colormaps <https://matplotlib.org/stable/tutorials/colors/colormaps.html>`_
            for supported colormaps.
        color_norm: str
            Set the 'lin' (linear) or 'log' (logarithmic) color normalization. If color_norm=None,
            the color normalization will be taken from the ASI array (log for black & white images 
            and linear for RGB images).
        pcolormesh_kwargs: dict
            Keyword arguments directly passed into plt.pcolormesh.
        """
        if ax is None:
            _, ax = plt.subplots()
        if (x_grid is None) and (y_grid is None):
            x_grid, y_grid = np.meshgrid(np.linspace(-12, 1.1, num=1000), np.linspace(-5, 5, num=1001))

        equator_sm, intensities = self.map_eq(b_model=b_model)

        gridded_eq_data = scipy.interpolate.griddata(
            equator_sm[:, :2], 
            intensities, 
            (x_grid, y_grid), 
            method='nearest'
            )
        # https://stackoverflow.com/a/31189177
        tree = cKDTree(equator_sm[:, :2])
        dists, _ = tree.query(np.stack((x_grid, y_grid), axis=-1))
        gridded_eq_data[dists > max_valid_grid_distance, :] = np.nan  # Mask any gridded point > 0.1 Re from the mapped point as NaN

        color_map, color_norm = self.imagers[0]._plot_params(gridded_eq_data, color_bounds, color_map, color_norm)

        vmin, vmax = self.imagers[0].plot_settings['color_bounds']
        if len(self.imagers[0].meta['resolution']) == 3:
            vmin, vmax = self.imagers[0].get_color_bounds()
            gridded_eq_data = self.imagers[0]._rgb_replacer(gridded_eq_data)
            gridded_eq_data = asilib.utils.stretch_contrast(gridded_eq_data, vmin, vmax)

            # Add a transprancy channel so that the NaN RGB values show up as transparent instead of black.
            gridded_eq_data = np.concatenate(
                (gridded_eq_data, np.ones(gridded_eq_data.shape[:2]).reshape(*gridded_eq_data.shape[:2], 1))
                , axis=-1
                )
            gridded_eq_data[np.isnan(gridded_eq_data.sum(axis=-1)), -1] = 0

        pcolormesh_kwargs['norm'] = color_norm
        pcolormesh_kwargs['cmap'] = color_map
        gridded_eq_data = np.ma.array(gridded_eq_data, mask=np.isnan(gridded_eq_data))
        p = ax.pcolormesh(x_grid, y_grid, gridded_eq_data, **pcolormesh_kwargs)

        # Draw Earth circle
        ax.add_artist(matplotlib.patches.Wedge((0, 0), 1, -90, 90, fc='w', ec='k'))
        ax.add_artist(matplotlib.patches.Wedge((0, 0), 1, 90, 270, fc='k', ec='k'))
        return ax, p

    def _igrf_model_wrapper(self, time, lla):
        if not hasattr(self, '_irbem_obj'):
            self._irbem_obj = IRBEM.MagFields(kext=0)
            self._coords_obj = IRBEM.Coords()
        output_dictionary = self._irbem_obj.find_magequator(
            {'datetime':time, 'x1':lla[2], 'x2':lla[0], 'x3':lla[1]},
            {}
            )
        equator_gsm = self._coords_obj.transform(time, output_dictionary['XGEO'], 1, 2)  # Convert to GSM coordinates
        return equator_gsm
    
    def __iter__(self) -> Generator[datetime, List, List]:
        """
        Generate a list of time stamps and images for all synchronized imagers, one time stamp 
        at a time.

        Yields
        ------
        datetime.datetime
            The guide time used to synchronize the imagers. If the difference between the imager
            time and the guide time is greater than time_tol*imager_cadence, or the imager is off,
            the imager is considered unsynchronized.
        list
            The time stamps from each imager. If the imager is unsynchronized the yielded 
            time stamp is ``datetime.min``.
        list 
            The images from each imager. If the imager is unsynchronized the yielded 
            image is ``None``.
        """
        guide_times = np.array([
            self.start_time+timedelta(seconds=i*self.min_cadence) for i in range(self.n_times)
            ])
        # asi_iterators keeps track of all ASIs in Imagers, with same order as passed into
        # __init__(). 
        # 
        # We must also keep track of ASIs whose next time stamp beyond the allowable tolerance 
        # of guide_time, or is off. future_iterator holds the imager (time, image) data until 
        # it becomes synchronized with the guide_time again.
        asi_iterators = {
            f'{_imager.meta["array"]}-{_imager.meta["location"]}':iter(_imager) 
            for _imager in self.imagers
            }
        
        current_images = {}
        future_iterators = {}
        stopped_iterators = []

        time_img_tuple = namedtuple('time_img_tuple', ['time', 'image'])

        for guide_time, next_guide_time in zip_longest(guide_times, guide_times[1:]):
            _asi_times = []
            _asi_images = []
            for (_name, _iterator), _imager in zip(asi_iterators.items(), self.imagers):
                try:
                    if _name in future_iterators:
                        current_images[_name] = future_iterators.pop(_name)
                    else:
                        asi_time, asi_img = next(_iterator)
                        current_images[_name] = time_img_tuple(asi_time, asi_img)
                except StopIteration:
                    _asi_times.append(datetime.min)
                    _asi_images.append(None)

                    if _name not in stopped_iterators:
                        stopped_iterators.append(_name)
                    if len(asi_iterators) == len(stopped_iterators):
                        # Stop once all iterators are exhausted.
                        return
                    continue


                # Check if the guide_time is within the cadence window after the current image 
                # timestamp, allowing for a fraction of an imager's cadence tolerance 
                # controlled by self.sync_image_tol. This addresses the case where the imager 
                # file for the minute 01:56:00 includes the first image at 01:56:00.1 
                # (which is within the tolerance).
                exposure_tol = timedelta(seconds=self.sync_image_tol*_imager.meta['cadence'])
                next_predicted_imager_timestamp = (
                    current_images[_name].time+timedelta(seconds=_imager.meta['cadence'])
                    )
                guide_time_after_image = (
                    guide_time >= (current_images[_name].time - exposure_tol)
                    )
                guide_time_before_next_image = (
                    guide_time <= next_predicted_imager_timestamp-exposure_tol
                    )

                if (guide_time_after_image and guide_time_before_next_image):
                    _asi_times.append(current_images[_name].time)
                    _asi_images.append(current_images[_name].image)

                    if (
                        (next_guide_time is not None) and 
                        (next_predicted_imager_timestamp-exposure_tol > next_guide_time)
                        ):
                        # Save the current image as a future image if the current image exposure takes 
                        # us beyond the next guide_time. This is useful for the slower of the imagers.
                        future_iterators[_name] = current_images[_name]
                else:
                    # Store the current image into the future iterator and 
                    # yield dummy values for time and image.
                    future_iterators[_name] = current_images[_name]
                    _asi_times.append(datetime.min)
                    _asi_images.append(None)


            if all(_asi_images is None for _asi_images in _asi_images):
                # This condition happens in the first iteration which has no images.
                # This is because the start_time is always earlier than all imagers' 
                # first image time.
                continue

            yield guide_time, _asi_times, _asi_images
        return

    def get_points(self, min_elevation:float=10)->Tuple[np.ndarray, np.ndarray]:
        """
        Get pixel intensities in each (lat, lon) grid point.

        Parameters
        ----------
        min_elevation: float
            Only return pixel intensities above min_elevation.

        Returns
        -------
        np.ndarray
            An (n, 2) array with each row corresponding to a (lat, lon) point.
        np.ndarray
            Pixel intensities with shape (n) for white-light images, and (n, 3) for RGB images.

        Examples
        --------
        >>> from datetime import datetime
        >>> 
        >>> import asilib
        >>> import asilib.asi
        >>> 
        >>> time = datetime(2007, 3, 13, 5, 8, 45)
        >>> location_codes = ['FSIM', 'ATHA', 'TPAS', 'SNKQ']
        >>> map_alt = 110
        >>> min_elevation = 2
        >>> 
        >>> _imagers = [asilib.asi.themis(location_code, time=time, alt=map_alt) 
        >>>             for location_code in location_codes]
        >>> asis = asilib.Imagers(_imagers)
        >>> lat_lon_points, intensities = asis.get_points(min_elevation=min_elevation)

        >>> # A comprehensive example showing how Imagers.get_points() can closely reproduce 
        >>> # Imagers.plot_map()
        >>> from datetime import datetime
        >>> 
        >>> import matplotlib.pyplot as plt
        >>> import matplotlib.colors
        >>> 
        >>> import asilib
        >>> import asilib.map
        >>> import asilib.asi
        >>> 
        >>> time = datetime(2007, 3, 13, 5, 8, 45)
        >>> location_codes = ['FSIM', 'ATHA', 'TPAS', 'SNKQ']
        >>> map_alt = 110
        >>> min_elevation = 2
        >>> 
        >>> _imagers = [asilib.asi.themis(location_code, time=time, alt=map_alt) 
        >>>             for location_code in location_codes]
        >>> asis = asilib.Imagers(_imagers)
        >>> lat_lon_points, intensities = asis.get_points(min_elevation=min_elevation)
        >>> 
        >>> fig = plt.figure(figsize=(12,5))
        >>> ax = asilib.map.create_simple_map(
        >>>     lon_bounds=(-140, -60), lat_bounds=(40, 82), fig_ax=(fig, 121)
        >>>     )
        >>> bx = asilib.map.create_simple_map(
        >>>     lon_bounds=(-140, -60), lat_bounds=(40, 82), fig_ax=(fig, 122)
        >>>     )
        >>> asis.plot_map(ax=ax, overlap=False, min_elevation=min_elevation)
        >>> bx.scatter(lat_lon_points[:, 1], lat_lon_points[:, 0], c=intensities, 
        >>>         norm=matplotlib.colors.LogNorm())
        >>> ax.text(0.01, 0.99, f'(A) Mosaic using Imagers.plot_map()', transform=ax.transAxes, 
        >>>         va='top', fontweight='bold', color='red')
        >>> bx.text(0.01, 0.99, f'(B) Mosaic from Imagers.get_points() scatter', transform=bx.transAxes,
        >>>         va='top', fontweight='bold', color='red')
        >>> fig.suptitle('Donovan et al. 2008 | First breakup of an auroral arc')
        >>> plt.tight_layout()
        >>> plt.show()
        """
        lat_lon_points = np.zeros((0, 2), dtype=float)
        if len(self.imagers[0].meta['resolution']) == 3: # RGB
            intensities  = np.zeros((0, self.imagers[0].meta['resolution'][-1]), dtype=float)
        else:  # single-color (or white light)
            intensities  = np.zeros(0, dtype=float)

        _skymaps = {}
        for imager in self.imagers:
            _skymaps[imager.meta['location']] = {
                'lon':imager.skymap['lon'].copy(), 
                'lat':imager.skymap['lat'].copy()
            }
        _skymaps = self.nan_overlap_pixels(_skymaps)

        for _imager in self.imagers:
            assert 'time' in _imager.file_info.keys(), (
                f'Imagers.get_points() only works with single images.'
                )
            _skymap_cleaner = Skymap_Cleaner(
                _skymaps[_imager.meta['location']]['lon'], 
                _skymaps[_imager.meta['location']]['lat'], 
                _imager.skymap['el'],
            )
            _masked_lon_map, _masked_lat_map, _ = _skymap_cleaner.mask_elevation(
                min_elevation=min_elevation
                )

            if _imager.skymap['lon'].shape[0] == _imager.data.image.shape[0] + 1:
                # Skymap defines vertices. We look for NaNs at either of the pixel edges.
                _valid_idx = np.where(
                    ~np.isnan(_masked_lon_map[1:, 1:]-_masked_lon_map[:-1, :-1])
                    )
            elif _imager.skymap['lon'].shape[0] == _imager.data.image.shape[0]:
                # Skymap defines pixel centers
                _valid_idx = np.where(~np.isnan(_masked_lon_map))
            else:
                raise ValueError(f'The skymap shape: {_imager.skymap["lon"].shape} and image '
                                 f'shape: {_imager.data.image.shape} are incompatible.')

            lat_grid = _masked_lat_map[_valid_idx[0], _valid_idx[1]]
            lon_grid = _masked_lon_map[_valid_idx[0], _valid_idx[1]]
            intensity = _imager.data.image[_valid_idx[0], _valid_idx[1], ...]

            # Concatenate joins arrays along an existing axis, while stack joins arrays
            # along a new axis. 
            intensities = np.concatenate((intensities, intensity))
            lat_lon_points = np.concatenate((
                lat_lon_points, 
                np.stack((lat_grid, lon_grid)).T
                ))
        return lat_lon_points, intensities
    
    def nan_overlap_pixels(
            self, 
            skymaps, 
            idx=None, 
            shrink_distance_to_self=0.98
            ):
        """
        Calculate which pixels whose mapped distance are closer to neighboring imagers and 
        NaN those. This is useful to remove the overlapping parts of the imagers when 
        plotting them together.

        Parameters
        ----------
        skymaps: dict
            A dictionary of (lat, lon) skymaps for each imager location. This is a copy of the 
            skymaps dictionary that this method modifies.
        idx: list
            A list of indices for self.imagers list to calculate the overlap mask for. 
            If None, the mask is calculated for all imagers.
        shrink_distance_to_self: float
            A scaling factor to apply to the distance from the pixels to its own imager. This
            is necessary to avoid gaps between the imager boundaries.

        Returns
        -------
        dict
            A dictionary of (lat, lon) skymaps in the same format as skyamps, for each imager location with the overlapping pixels masked as np.nan.
        """
        if idx is None:
            _imagers = self.imagers
        else:
            _imagers = [self.imagers[i] for i in idx]

        for i, imager in enumerate(_imagers):
            # Algorithm:
            # 1. Loop over ith imager
            # 2. Loop over jth imager within 500 km distance of the ith imager
            # 3. Mask low-elevations with np.nan
            # 4. Create a distance array with shape (resolution[0], resolution[1], j_total)
            # 5. For all pixels in ith imager, calculate the haversine distance to jth imager and 
            # assign it to distance[..., j].
            # 6. For all pixels calculate the nearest imager out of all j's.
            # 7. If the minimum j is not the ith imager, mask the imager.skymap['lat'] and 
            # imager.skymap['lon'] as np.nan.
            _distances = np.nan*np.ones(
                (*skymaps[imager.meta['location']]['lat'].shape, len(_imagers))
                )
            for j, other_imager in enumerate(_imagers):
                # Calculate the distance between all imager pixels and every other imager 
                # location (including itself).
                _other_lon = np.broadcast_to(
                    other_imager.meta['lon'], 
                    skymaps[imager.meta['location']]['lat'].shape
                    )
                _other_lat = np.broadcast_to(
                    other_imager.meta['lat'], 
                    skymaps[imager.meta['location']]['lat'].shape
                    )

                _distances[:, :, j] = _haversine(
                    skymaps[imager.meta['location']]['lat'], 
                    skymaps[imager.meta['location']]['lon'],
                    _other_lat, 
                    _other_lon
                    )
                
            # Without this small reduction in the distance of pixels to its own imager,
            # there are gaps between the imager boundaries. In other words, this scaling
            # slightly biases to plotting pixels nearest to the imager. 
            _distances[:, :, i] *= shrink_distance_to_self
            # TODO: Set the _distances[:, :, i] values to 0 (not masked out) if they are 
            # #closer to the other imager but the other imager's pixels are nans.
            # Need a masked array so that np.nanargmin correctly handles all NaN slices.
            _distances = np.ma.masked_array(_distances, np.isnan(_distances))
            # For each pixel, calculate the nearest imager. If the pixel is not closest to 
            # the imager that it's from, mask it as np.nan.
            min_distances = np.argmin(_distances, axis=2)
            far_pixels = np.where(min_distances != i)

            skymaps[imager.meta['location']]['lat'][far_pixels] = np.nan
            skymaps[imager.meta['location']]['lon'][far_pixels] = np.nan
        return skymaps

    def find_overlap_pixels(self, idx=None, min_elevation:float=10) -> dict:
        """
        Similar to nan_overlap_pixels, but instead of masking the overlapping pixels as np.nan, 
        it returns a dictionary of dictionaties containing indices arrays of the overlapping 
        pixels for each imager and a neighboring imager. This is useful for triangulation.

        Parameters
        ----------
        idx: list
            A list of indices for self.imagers list to calculate the overlap pixels for. 
            If None, the pixels are calculated for all imagers.
        min_elevation: float
            Only consider pixels above min_elevation when calculating the overlapping pixels.

        Returns
        -------
        dict
            A dictionary of dictionaries containing indices arrays of the overlapping pixels 
            for each imager and a neighboring imager. The keys of the outer dictionary are the 
            self imager, and the inner dictionary has keys of the neighboring imager with thich
            the self imager has overlapping pixels.

        Examples
        --------
        >>> import asilib.asi
        >>> import asilib.map
        >>> import matplotlib.pyplot as plt
        >>> import cartopy.crs as ccrs
        >>> 
        >>> time = datetime(2007, 3, 13, 5, 8, 45)
        >>> location_codes = ['ATHA', 'TPAS']
        >>> map_alt = 110
        >>> _imagers = [asilib.asi.themis(location_code, time=time, alt=map_alt) 
        >>>             for location_code in location_codes]
        >>> asis = asilib.Imagers(_imagers)
        >>> overlapping_masks = asis.find_overlap_pixels(min_elevation=2)
        >>> 
        >>> ax = asilib.map.create_map(
        >>>     lon_bounds=(-125, -90), lat_bounds=(42, 64)
        >>>     )            
        >>> 
        >>> asis.plot_map(ax=ax, min_elevation=2)
        >>> 
        >>> asi_names = [f'{_imager.meta["array"]}-{_imager.meta["location"]}' for _imager in asis.imagers]
        >>> 
        >>> for self_loc, neighbors in overlapping_masks.items():
        >>>     for neighbor_loc, overlapping_mask in neighbors.items():
        >>>         self_loc_idx = asi_names.index(self_loc)
        >>>         neighbor_loc_idx = asi_names.index(neighbor_loc)
        >>>         ax.scatter(
        >>>             asis.imagers[self_loc_idx].skymap['lon'][overlapping_mask], 
        >>>             asis.imagers[self_loc_idx].skymap['lat'][overlapping_mask], 
        >>>             alpha=0.5, 
        >>>             label=f'{self_loc} pixels overlapping with {neighbor_loc}',
        >>>             s=2,
        >>>             transform=ccrs.PlateCarree()
        >>>         )   
        >>>         
        >>> ax.legend(loc='lower left', fontsize=12, markerscale=3)
        >>> plt.tight_layout()
        >>> plt.show()
        """
        overlap_pixels = {}
        if idx is None:
            _imagers = self.imagers
        else:
            _imagers = [self.imagers[i] for i in idx]

        for i, self_imager in enumerate(_imagers):
            # Algorithm:
            # 1. Loop over ith imager
            # 2. Loop over jth imager within 500 km distance of the ith imager
            # 3. Mask low-elevations with np.nan
            # 4. Create a distance array with shape (resolution[0], resolution[1], j_total)
            # 5. For all pixels in ith imager, calculate the haversine distance to jth imager and 
            # assign it to distance[..., j].
            # 6. For all pixels calculate the nearest imager out of all j's.
            # 7. If the minimum j is not the ith imager, add the pixel indices to the 
            # overlap_pixels dictionary with outer key of the self imager and the inner 
            # key neighboring imager.
            _distances = np.nan*np.ones(
                (*self_imager.skymap['lat'].shape, len(_imagers))
                )
            
            _skymap_cleaner = Skymap_Cleaner(
                self_imager.skymap['lon'], 
                self_imager.skymap['lat'], 
                self_imager.skymap['el'], 
            )
            _skymap_cleaner.mask_elevation(min_elevation)
                
            for j, other_imager in enumerate(_imagers):
                # Calculate the distance between all imager pixels and every other imager 
                # location (including itself).
                _other_lon = np.broadcast_to(
                    other_imager.meta['lon'], 
                    _skymap_cleaner._lat_grid.shape
                    )
                _other_lat = np.broadcast_to(
                    other_imager.meta['lat'], 
                    _skymap_cleaner._lat_grid.shape
                    )

                _distances[:, :, j] = _haversine(
                    _skymap_cleaner._lat_grid, 
                    _skymap_cleaner._lon_grid,
                    _other_lat, 
                    _other_lon
                    )
                
            _distances = np.ma.masked_array(_distances, np.isnan(_distances))
            # For each pixel, calculate the nearest imager. If the pixel is not closest to 
            # the imager that it's from, add its indices to the overlap_pixels dictionary.
            min_distances = np.argmin(_distances, axis=2)

            for j, other_imager in enumerate(_imagers):
                ij_overlap_pixels = np.where(
                    (min_distances == j) & 
                    (j != i) & 
                    ~np.isnan(_skymap_cleaner._lat_grid)
                    )

                self_key =f'{self_imager.meta["array"]}-{self_imager.meta["location"]}'
                other_key = f'{other_imager.meta["array"]}-{other_imager.meta["location"]}'

                if ij_overlap_pixels[0].shape[0] > 0:
                    if self_key not in overlap_pixels.keys():
                        overlap_pixels[self_key] = {}
                    overlap_mask = np.zeros(self_imager.skymap['lat'].shape, dtype=bool)
                    overlap_mask[ij_overlap_pixels] = True
                    overlap_pixels[self_key][other_key] = overlap_mask
        return overlap_pixels
    
    def __str__(self):
        names = [f'{_img.meta["array"]}-{_img.meta["location"]}' for _img in self.imagers]
        names = 'asilib.Imagers initiated with:\n' + ', '.join(names)

        if (
                ('time' in self.imagers[0].file_info.keys()) and 
                (self.imagers[0].file_info['time'] is not None)
            ):
            return (names + f'\ntime={self.imagers[0].file_info["time"]}')
        elif (
                ('time_range' in self.imagers[0].file_info.keys()) and 
                (self.imagers[0].file_info['time_range'] is not None)
            ):
            return (names + f'\ntime_range={self.imagers[0].file_info["time_range"]}')
        else:
            raise ValueError(
                'The 0th imager object does not have a "time" or a "time_range" variable.'
                )
        

if __name__ == '__main__':
    import asilib.asi
    import asilib.map
    import matplotlib.pyplot as plt
    import cartopy.crs as ccrs

    time = datetime(2007, 3, 13, 5, 8, 45)
    location_codes = ['ATHA', 'TPAS']
    map_alt = 110
    _imagers = [asilib.asi.themis(location_code, time=time, alt=map_alt) 
                for location_code in location_codes]
    asis = asilib.Imagers(_imagers)
    overlapping_masks = asis.find_overlap_pixels(min_elevation=2)

    ax = asilib.map.create_map(
        lon_bounds=(-125, -90), lat_bounds=(42, 64)
        )            

    asis.plot_map(ax=ax, min_elevation=2)

    asi_names = [f'{_imager.meta["array"]}-{_imager.meta["location"]}' for _imager in asis.imagers]

    for self_loc, neighbors in overlapping_masks.items():
        for neighbor_loc, overlapping_mask in neighbors.items():
            self_loc_idx = asi_names.index(self_loc)
            neighbor_loc_idx = asi_names.index(neighbor_loc)
            ax.scatter(
                asis.imagers[self_loc_idx].skymap['lon'][overlapping_mask], 
                asis.imagers[self_loc_idx].skymap['lat'][overlapping_mask], 
                alpha=0.5, 
                label=f'{self_loc} pixels overlapping with {neighbor_loc}',
                s=2,
                transform=ccrs.PlateCarree()
            )   
            
    ax.legend(loc='lower left', fontsize=12, markerscale=3)
    plt.tight_layout()
    plt.show()

    # # You will need to install cdasws to run this example (python -m pip install cdasws)
    # from datetime import datetime, timedelta, timezone
    
    # import cdasws
    # import matplotlib.pyplot as plt
    # import matplotlib.dates
    # import pandas as pd
    # import asilib.asi
    
    # time_range=(datetime(2021, 11, 4, 1, 56), datetime(2021, 11, 4, 12, 30))  #datetime(2021, 11, 4, 12, 24)
    # mango_location_code='CFS'
    # mango_asi = asilib.asi.mango(mango_location_code, 'redline', time_range=time_range)
    # trex_location_codes = ['FSMI', 'LUCK', 'RABB', 'PINA', 'GILL']
    # trex_asis = [asilib.asi.trex_rgb(location_code, time_range=time_range) 
    #               for location_code in trex_location_codes]
    # asis = asilib.Imagers([mango_asi]+trex_asis)

    # fig = plt.figure(layout='constrained', figsize=(8, 9))
    # gs = matplotlib.gridspec.GridSpec(2, 1, fig, height_ratios=(3, 1))
    # ax = asilib.map.create_map(lat_bounds=(30, 64), lon_bounds=(-125, -75), fig_ax=(fig, gs[0]))
    # bx = fig.add_subplot(gs[1])
    # bx.xaxis.set_major_formatter(matplotlib.dates.DateFormatter('%H:%M'))
    # gen = asis.animate_map_gen(ax=ax, asi_label=True, overwrite=True, ffmpeg_params={'framerate':100})
    
    # cdas = cdasws.CdasWs()
    # time_range_cdasws = cdasws.TimeInterval(
    #     datetime.fromisoformat(str(time_range[0]-timedelta(days=0.5))).replace(tzinfo=timezone.utc),
    #     datetime.fromisoformat(str(time_range[1]+timedelta(days=0.5))).replace(tzinfo=timezone.utc)
    #     )
    # _, data = cdas.get_data(
    #                 'OMNI_HRO_5MIN', ['SYM_H'], time_range_cdasws
    #                 )
    # symh = pd.DataFrame(index=data['SYM_H'].Epoch.data, data={'SYM_H':data['SYM_H']})
    # bx.plot(symh.index, symh['SYM_H'], c='k')
    # bx.set(xlabel='Time [HH:MM]', ylabel='Sym-H [nT]')

    # plt.suptitle(f'MANGO & TREx-RGB Mosaic | {time_range[0]:%F %T} to {time_range[1]:%T}', fontsize=16)
    
    # for guide_time, image_times, images, ax in gen:
    #     # We will need to delete the prior text object, otherwise the current one
    #     # will overplot on the prior one---clean up after yourself!
    #     if '_time_guide' in locals():
    #         _time_guide.remove()  # noqa: F821
    #         _text_box.remove()  # noqa: F821
    #     _zip = zip([mango_location_code]+trex_location_codes, image_times)
    #     time_str = (
    #         f'Current timestamps:\n'+
    #         '\n'.join([f'{name} {ti.strftime("%H:%M:%S.%f")}' for name, ti in _zip])
    #         )
    #     _text_box = ax.text(
    #         0.75, 
    #         0.01, 
    #         time_str, 
    #         color='w',
    #         transform=ax.transAxes, 
    #         bbox=dict(boxstyle='round', facecolor='purple', alpha=0.5), 
    #         ha='left', 
    #         va='bottom'
    #         )
    #     _time_guide = bx.axvline(guide_time, c='k', ls='--')
        


"""
This example script creates a movie with the satellite track superposed.
"""
from datetime import datetime

import numpy as np
import matplotlib.pyplot as plt

import asilib
import asilib.asi


# ASI parameters
location_code = 'RANK'
alt = 110  # km
time_range = (datetime(2017, 9, 15, 2, 32, 0), datetime(2017, 9, 15, 2, 35, 0))

fig, ax = plt.subplots(
    3, 1, figsize=(7, 10), gridspec_kw={'height_ratios': [4, 1, 1]}, constrained_layout=True
)

asi = asilib.asi.themis(location_code, time_range=time_range, alt=alt)

# Create the fake satellite track coordinates: latitude, longitude, altitude (LLA).
# This is a north-south satellite track oriented to the east of the THEMIS/RANK
# imager.
n = int((time_range[1] - time_range[0]).total_seconds() / 3)  # 3 second cadence.
lats = np.linspace(asi.meta["lat"] + 5, asi.meta["lat"] - 5, n)
lons = (asi.meta["lon"] - 0.5) * np.ones(n)
alts = alt * np.ones(n)  # Altitude needs to be the same as the skymap.
sat_lla = np.array([lats, lons, alts]).T
# Normally the satellite time stamps are not the same as the ASI.
# You may need to call Conjunction.interp_sat() to find the LLA coordinates
# at the ASI timestamps.
sat_time = asi.data.time

conjunction_obj = asilib.Conjunction(asi, (sat_time, sat_lla))

# Map the satellite track to the imager's azimuth and elevation coordinates and
# image pixels. NOTE: the mapping is not along the magnetic field lines! You need
# to install IRBEM and then use conjunction.lla_footprint() before
# calling conjunction_obj.map_azel.
sat_azel, sat_azel_pixels = conjunction_obj.map_azel()

# Calculate the auroral intensity near the satellite and mean intensity within a 10x10 km area.
nearest_pixel_intensity = conjunction_obj.intensity(box=None)
area_intensity = conjunction_obj.intensity(box=(10, 10))
area_mask = conjunction_obj.equal_area(box=(10, 10))

# Need to change masked NaNs to 0s so we can plot the rectangular area contours.
area_mask[np.where(np.isnan(area_mask))] = 0

# Initiate the animation generator function.
gen = asi.animate_fisheye_gen(
    ax=ax[0], azel_contours=True, overwrite=True, cardinal_directions='NE'
)

for i, (time, image, _, im) in enumerate(gen):
    # Plot the entire satellite track, its current location, and a 20x20 km box
    # around its location.
    ax[0].plot(sat_azel_pixels[:, 0], sat_azel_pixels[:, 1], 'red')
    ax[0].scatter(sat_azel_pixels[i, 0], sat_azel_pixels[i, 1], c='red', marker='o', s=50)
    ax[0].contour(area_mask[i, :, :], levels=[0.99], colors=['yellow'])

    if 'vline1' in locals():
        vline1.remove()  # noqa: F821
        vline2.remove()  # noqa: F821
        text_obj.remove()  # noqa: F821
    else:
        # Plot the ASI intensity along the satellite path
        ax[1].plot(sat_time, nearest_pixel_intensity)
        ax[2].plot(sat_time, area_intensity)
    vline1 = ax[1].axvline(time, c='b')
    vline2 = ax[2].axvline(time, c='b')

    # Annotate the location_code and satellite info in the top-left corner.
    location_code_str = (
        f'THEMIS/{location_code} '
        f'LLA=({asi.meta["lat"]:.2f}, '
        f'{asi.meta["lon"]:.2f}, {asi.meta["alt"]:.2f})'
    )
    satellite_str = f'Satellite LLA=({sat_lla[i, 0]:.2f}, {sat_lla[i, 1]:.2f}, {sat_lla[i, 2]:.2f})'
    text_obj = ax[0].text(
        0,
        1,
        location_code_str + '\n' + satellite_str,
        va='top',
        transform=ax[0].transAxes,
        color='red',
    )
    ax[1].set(ylabel='ASI intensity\nnearest pixel [counts]')
    ax[2].set(xlabel='Time', ylabel='ASI intensity\n10x10 km area [counts]')

print(f'Animation saved in {asilib.config["ASI_DATA_DIR"] / "animations" / asi.animation_name}')



from datetime import datetime

import pytest
import numpy as np
import matplotlib.testing.decorators

import asilib
import asilib.asi
from asilib.asi.fake_asi import fake_asi


##########################################
########## TEST DATA ACCESS ##############
##########################################
def test_time():
    """
    Test if the fake_asi timestamp is correctly accessed.
    """
    asi = fake_asi('GILL', time='2015-01-01T15:14:00.17')
    assert asi.data.time == datetime(2015, 1, 1, 15, 14)
    assert asi.data.image.shape == (512, 512)
    assert np.isclose(asi.data.image.mean(), 14.4014892578125)
    # See https://numpy.org/doc/stable/reference/generated/numpy.argmax.html for the
    # unravel_index example to get the maximum index for a N-d array.
    ind = np.unravel_index(np.argmax(asi.data.image, axis=None), asi.data.image.shape)
    assert ind == (318, 0)
    return


def test_time_range():
    """
    Test if fake_asi timestamps are correctly accessed.
    """
    asi = fake_asi('GILL', time_range=('2015-01-01T15:00:15.17', '2015-01-01T20:00'))
    assert asi.data.time.shape == (1798,)
    assert asi.data.image.shape == (1798, 512, 512)
    assert np.isclose(asi.data.image.mean(), 14.048684923216552)
    assert asi.file_info['path'] == [
        '20150101_150000_GILL_fake_asi.png',
        '20150101_160000_GILL_fake_asi.png',
        '20150101_170000_GILL_fake_asi.png',
        '20150101_180000_GILL_fake_asi.png',
        '20150101_190000_GILL_fake_asi.png',
    ]
    return


def test_iter_files():
    """
    Test Imager.iter_files()
    """
    asi = fake_asi('GILL', time_range=('2015-01-01T15:00:15.17', '2015-01-01T19:30'))
    n_timestamps_reference = [358, 360, 360, 360, 181]

    counter = 0
    for (_times, _images), ref_n in zip(asi.iter_files(), n_timestamps_reference):
        assert _times.shape[0] == ref_n
        assert _images.shape[0] == ref_n
        counter += 1

    assert counter == len(n_timestamps_reference)
    return


##########################################
############# TEST PLOTTING ##############
##########################################
@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_fisheye'], tol=10, remove_text=True, extensions=['png']
)
def test_fisheye():
    """
    Tests that the fake_asi produces the correct image.
    """
    asi = fake_asi('GILL', time='2015-01-01T15:14:00.17')
    asi.plot_fisheye(color_bounds=(1, 255), origin=(0.85, 0.15), cardinal_directions='NEWS')
    return


@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_plot_keogram'], tol=10, remove_text=True, extensions=['png']
)
def test_plot_keogram():
    """
    Tests that the fake_asi produces the correct keogram.
    """
    asi = fake_asi('GILL', time_range=('2015-01-01T15:00:15.17', '2015-01-01T19:30'))
    asi.plot_keogram()
    return


##########################################
############# TEST EXAMPLES ##############
##########################################


@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_plot_fisheye_example'], tol=10, remove_text=True, extensions=['png']
)
def test_plot_fisheye_example():
    """
    Test that asilib.Imager plots a bright auroral arc that was analyzed by
    Imajo et al., 2021 "Active auroral arc powered by accelerated electrons
    from very high altitudes"
    """
    from datetime import datetime
    import matplotlib.pyplot as plt
    import asilib.asi

    asi = asilib.asi.themis('RANK', time=datetime(2017, 9, 15, 2, 34, 0))
    ax, im = asi.plot_fisheye(cardinal_directions='NE', origin=(0.95, 0.05))
    plt.colorbar(im)
    ax.axis('off')


@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_plot_map_example'], tol=10, remove_text=True, extensions=['png']
)
def test_plot_map_example():
    """
    Test a mapped image of STEVE.
    """
    # Project an image of STEVE onto a map.
    from datetime import datetime
    import matplotlib.pyplot as plt
    import asilib.asi
    import asilib.map

    asi = asilib.asi.themis('ATHA', time=datetime(2010, 4, 5, 6, 7, 0))
    ax = asilib.map.create_simple_map(lon_bounds=(-127, -100), lat_bounds=(45, 65))
    asi.plot_map(ax=ax)
    plt.tight_layout()
    return


def test_animate_fisheye_example():
    """
    Test that asilib.Imager.animate_fisheye() method works. Since I have not
    found a way to compare animations, lets assume that the plot_fisheye()
    tests in the underlying plotting methods pass, and we just need
    to check that the movie exists with the correct number of underlying plots.
    """
    from datetime import datetime
    import asilib.asi

    time_range = (datetime(2015, 3, 26, 6, 7), datetime(2015, 3, 26, 6, 12))
    asi = asilib.asi.themis('FSMI', time_range=time_range)
    asi.animate_fisheye(cardinal_directions='NE', origin=(0.95, 0.05), overwrite=True)
    print(f'Animation saved in {asilib.config["ASI_DATA_DIR"] / "animations" / asi.animation_name}')

    # End of example.
    animation_path = asilib.config["ASI_DATA_DIR"] / "animations" / asi.animation_name
    image_parent_dir_chunks = asi.animation_name.split('_')
    image_parent_dir = '_'.join(image_parent_dir_chunks[0:2]) + '_'
    image_parent_dir = image_parent_dir + '_'.join(image_parent_dir_chunks[3:])
    image_parent_dir = image_parent_dir.split('.')[0]
    plot_dir = asilib.config["ASI_DATA_DIR"] / "animations" / 'images' / image_parent_dir
    image_paths = list(plot_dir.glob('*.png'))

    assert animation_path.exists()
    assert plot_dir.exists()
    assert len(image_paths) == 100
    return


def test_animate_map_example():
    """
    Test that asilib.Imager.animate_map() method works. Since I have not
    found a way to compare animations, lets assume that the plot_map()
    tests in the underlying plotting methods pass, and we just need
    to check that the movie exists with the correct number of underlying plots.
    """
    from datetime import datetime
    import asilib.asi
    import asilib.map
    import asilib

    location = 'FSMI'
    time_range = (datetime(2015, 3, 26, 6, 7), datetime(2015, 3, 26, 6, 8, 30))
    asi = asilib.asi.themis(location, time_range=time_range)
    ax = asilib.map.create_simple_map()
    asi.animate_map(overwrite=True, ax=ax)
    print(f'Animation saved in {asilib.config["ASI_DATA_DIR"] / "animations" / asi.animation_name}')

    # End of example.
    animation_path = asilib.config["ASI_DATA_DIR"] / "animations" / asi.animation_name
    image_parent_dir_chunks = asi.animation_name.split('_')
    image_parent_dir = '_'.join(image_parent_dir_chunks[0:2]) + '_'
    image_parent_dir = image_parent_dir + '_'.join(image_parent_dir_chunks[3:])
    image_parent_dir = image_parent_dir.split('.')[0]
    plot_dir = asilib.config["ASI_DATA_DIR"] / "animations" / 'images' / image_parent_dir
    image_paths = list(plot_dir.glob('*.png'))

    assert animation_path.exists()
    assert plot_dir.exists()
    assert len(image_paths) == 30
    return


@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_plot_keogram_geographic_example'],
    tol=10,
    remove_text=True,
    extensions=['png'],
)
def test_plot_keogram_geographic_example():
    """
    Tests that the keogram is correctly plotted with geographic lats
    Event from https://doi.org/10.1029/2021GL094696
    """
    import matplotlib.pyplot as plt
    import asilib.asi

    time_range = ['2008-01-16T10', '2008-01-16T12']
    asi = asilib.asi.themis('GILL', time_range=time_range)
    ax, p = asi.plot_keogram(color_map='turbo')
    ax.set_ylabel('Geographic Lat [deg]')
    plt.colorbar(p)
    return


@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_plot_keogram_magnetic_example'],
    tol=10,
    remove_text=True,
    extensions=['png'],
)
def test_plot_keogram_magnetic_example():
    """
    Tests that the keogram is correctly plotted with magnetic lats
    Event from https://doi.org/10.1029/2021GL094696
    """
    import matplotlib.pyplot as plt
    import asilib.asi

    time_range = ['2008-01-16T10', '2008-01-16T12']

    asi = asilib.asi.themis('GILL', time_range=time_range)
    ax, p = asi.plot_keogram(
        color_map='turbo', aacgm=True, title=False,
        color_bounds=asi.auto_color_bounds()
        )
    ax.set_ylabel('Magnetic Lat [deg]')
    plt.colorbar(p)
    return


@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_plot_keogram_magnetic_path_example'],
    tol=10,
    remove_text=True,
    extensions=['png'],
)
def test_plot_keogram_magnetic_path_example():
    """
    Tests that the keogram is correctly plotted with magnetic lats
    Event from https://doi.org/10.1029/2021GL094696
    """
    import matplotlib.pyplot as plt
    import numpy as np
    import asilib.asi

    time_range = ['2008-01-16T10', '2008-01-16T12']

    asi = asilib.asi.themis('GILL', time_range=time_range)

    latlon = np.column_stack(
        (
            asi.skymap['lat'][:, asi.meta['resolution'][0] // 3],
            asi.skymap['lon'][:, asi.meta['resolution'][0] // 3],
        )
    )
    latlon = latlon[np.where(~np.isnan(latlon[:, 0]))[0], :]
    ax, p = asi.plot_keogram(
        color_map='turbo', aacgm=True, title=False, path=latlon, color_bounds=(4034, 6886)
        )
    ax.set_ylabel('Magnetic Lat [deg]\nCustom path')
    plt.colorbar(p)
    return

def test_get_set_color_bounds():
    """
    Test that we can get and set the color bounds.
    """
    import matplotlib.pyplot as plt
    import numpy as np
    import asilib.asi

    time_range = ['2008-01-16T10', '2008-01-16T12']

    asi = asilib.asi.themis('GILL', time_range=time_range)
    assert np.all(np.array(asi.auto_color_bounds()).round() == np.array([4034., 6887.]))
    asi.set_color_bounds(4000, 7000)
    assert np.all(asi.get_color_bounds() == (4000, 7000))
    return


def test_getitem():
    """
    Tests the __getitem__() for time slicing.
    """
    import asilib.asi

    time_range = ['2008-01-16T10', '2008-01-16T12']

    asi = asilib.asi.themis('GILL', time_range=time_range)

    asi_one_time = asi['2008-01-16T11:05:10']
    assert asi_one_time.data.time == datetime(2008, 1, 16, 11, 5, 9, 10571)
    assert np.all(
        asi_one_time.data.image[0, :10]
        == np.array([3498, 3527, 3459, 3516, 3484, 3505, 3483, 3479, 3499, 3481])
    )
    with pytest.raises(FileNotFoundError):
        # Outside time_range.
        asi['2010-01-16T11:00:00']
        asi['2005-01-16T11:00:00']
    asi_time_range = asi['2008-01-16T10:30' : datetime(2008, 1, 16, 10, 40)]
    assert asi_time_range.data.time.shape[0] == 200
    assert asi_time_range.data.time[0] == datetime(2008, 1, 16, 10, 30, 0, 16380)
    assert asi_time_range.data.time[-1] == datetime(2008, 1, 16, 10, 39, 57, 25495)
    assert asi_time_range.data.image.shape == (200, 256, 256)
    return


def test_str():
    """
    Tests the __str__() for printing user-readable information about the imager
    """
    import asilib.asi

    time_range = ['2008-01-16T10', '2008-01-16T12']

    asi = asilib.asi.themis('GILL', time_range=time_range)
    assert str(asi) == (
        'A THEMIS-GILL Imager. time_range=[datetime.datetime(2008, 1, 16, 10, 0), '
        'datetime.datetime(2008, 1, 16, 12, 0)]'
    )

    asi2 = asilib.asi.themis('GILL', time=time_range[0])
    assert str(asi2) == 'A THEMIS-GILL Imager. time=2008-01-16 10:00:00'
    return


def test_repr():
    """
    Tests the __repr__() for printing machine-readable information about the imager
    """
    asi = asilib.Imager(
        file_info={'test_variable': 0},
        skymap={'skymap': 1},
        meta={'meta': 2},
        plot_settings={'plot_settings': 3},
    )
    assert repr(asi) == (
        "Imager(file_info={'test_variable': 0}, skymap={'skymap': 1}, meta={'meta': 2},"
        " plot_settings={'plot_settings': 3})"
    )
    return



"""
Tests the themis() data loading and the example plotting functions.
"""
from datetime import datetime

import requests
import pytest
import matplotlib.testing.decorators
import numpy as np

import asilib.asi

##########################################
############# TEST LOADERS ###############
##########################################


def test_themis_time():
    """
    Tests that one file is loaded and themis() returns the correct file.
    """
    # Calls the download function
    asi = asilib.asi.themis('gill', time='2014-05-05T04:49:10', redownload=True)
    # asi.file_info should not be accessed by the user.
    assert asi.file_info['time'] == datetime(2014, 5, 5, 4, 49, 10)
    # But asi.data should be
    assert asi.data.time == datetime(2014, 5, 5, 4, 49, 9, 37070)
    assert asi.data[0] == datetime(2014, 5, 5, 4, 49, 9, 37070)
    assert asi.data.image[0, 0] == 3464
    assert asi.data[1][0, 0] == 3464
    assert asi.data.image[-1, -1] == 3477
    assert asi.skymap['path'].name == 'themis_skymap_gill_20130103-%2B_vXX.sav'
    # Does not call the download function
    asi2 = asilib.asi.themis('gill', time='2014-05-05T04:49:10', redownload=False)
    assert asi2.file_info['time'] == datetime(2014, 5, 5, 4, 49, 10)
    return


def test_themis_time_range():
    """
    Tests that multiple files are loaded using themis()'s time_range kwarg.
    """
    img = asilib.asi.themis('gill', time_range=['2014-05-05T05:10', '2014-05-05T05:12'])
    assert img.file_info['path'][0].name == '20140505_0510_gill_themis19_full.pgm.gz'
    assert img.file_info['path'][1].name == '20140505_0511_gill_themis19_full.pgm.gz'

    assert img.data.time[0] == datetime(2014, 5, 5, 5, 10, 0, 30996)
    assert img.data.time[-1] == datetime(2014, 5, 5, 5, 11, 57, 23046)
    assert img.data.time.shape == (40,)
    assert img.data.image.shape == (40, 256, 256)
    return


def test_themis_no_file():
    """
    Tests that themis() returns a FileNotFound error when we try to download
    non-existant data.
    """
    with pytest.raises(requests.exceptions.HTTPError):
        asilib.asi.themis('pina', time='2011/07/07T02:00')
    return


def test_themis_partial_files():
    """
    Tests that themis() returns a FileNotFound error when we try to download
    non-existant data.
    """
    with pytest.raises(FileNotFoundError):
        asilib.asi.themis(
            'pina', time_range=['2011/07/07T04:15', '2011/07/07T04:23'], missing_ok=False
        )

    img = asilib.asi.themis(
        'pina',
        time_range=['2011/07/07T04:15', '2011/07/07T04:23:30'],
        missing_ok=True,
        redownload=False,
    )
    assert img.file_info['path'][0].name == '20110707_0421_pina_themis18_full.pgm.gz'
    assert img.file_info['path'][1].name == '20110707_0422_pina_themis18_full.pgm.gz'
    assert img.file_info['path'][2].name == '20110707_0423_pina_themis18_full.pgm.gz'
    return


def test_themis_asi_meta():
    """
    Checks that the THEMIS ASI metadata is correct.
    """
    img = asilib.asi.themis(
        'pina',
        time_range=['2011/07/07T04:20', '2011/07/07T04:22:00'],
        missing_ok=True,
        redownload=False,
    )
    assert img.meta == {
        'array': 'THEMIS',
        'location': 'PINA',
        'lat': 50.15999984741211,
        'lon': -96.07000732421875,
        'alt': 0.0,
        'cadence': 3,
        'resolution': (256, 256),
    }
    return


##########################################
############# TEST EXAMPLES ##############
##########################################
@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_themis_fisheye'], tol=10, remove_text=True, extensions=['png']
)
def test_themis_fisheye():
    """
    Plot a fisheye lens image.
    """
    from datetime import datetime
    import matplotlib.pyplot as plt
    import asilib.asi

    asi = asilib.asi.themis('RANK', time=datetime(2017, 9, 15, 2, 34, 0))
    ax, im = asi.plot_fisheye(cardinal_directions='NE', origin=(0.95, 0.05))
    plt.colorbar(im)
    ax.axis('off')
    return

def test_themis_download_bug_fix():
    """
    Test a bug if an imager does not have data for that hour, and that if missing_ok is set,
    the Imager.keogram() raises a ValueError.
    """
    location_code = 'KAPU'
    time_range = (datetime(2013, 3, 17, 11), datetime(2013, 3, 17, 11, 1))

    with pytest.warns(UserWarning) as warninfo:
        asi = asilib.asi.themis(location_code, time_range=time_range)

    with pytest.raises(ValueError) as excinfo:
        asi.plot_keogram()
    
    # check that only one warning was raised
    assert len(warninfo) == 1
    # check that the message matches
    assert 'No local or online image files found' in warninfo[0].message.args[0]
    assert 'The keogram is empty' in str(excinfo.value)
    return


@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_themis_map'], tol=10, remove_text=True, extensions=['png']
)
def test_themis_map():
    """
    Plot a fisheye lens image on a map.
    """
    from datetime import datetime
    import matplotlib.pyplot as plt
    import asilib.asi
    import asilib.map

    asi = asilib.asi.themis('RANK', time=datetime(2017, 9, 15, 2, 34, 0))
    ax = asilib.map.create_simple_map()
    ax, im = asi.plot_map(ax=ax)
    plt.colorbar(im)
    ax.axis('off')
    return


@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_themis_keogram'], tol=10, remove_text=True, extensions=['png']
)
def test_themis_keogram():
    """
    Plot a keogram.
    """
    from datetime import datetime
    import matplotlib.pyplot as plt
    import asilib.asi

    asi = asilib.asi.themis(
        'RANK', time_range=(datetime(2017, 9, 15, 2, 30, 0), datetime(2017, 9, 15, 2, 35, 0))
    )
    ax, im = asi.plot_keogram()
    plt.colorbar(im)
    ax.axis('off')
    return

def test_themis_available_time():
    time = '2018-01-31T14:00'
    availability = asilib.asi.themis_available(time=time)
    assert sum(availability) == 9
    return

def test_themis_available_time_range():
    time_range = ('2018-01-31T13:01:27', '2018-01-31T20:00')
    availability = asilib.asi.themis_available(time_range=time_range)
    assert np.all(~availability['YKNF']) is np.True_
    assert sum(availability.loc['2018-01-31T14:00', :]) == 9
    assert sum(availability.loc['2018-01-31T20:00', :]) == 0
    assert availability.shape == (8, 23)
    return

@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_themis_available_multiday_time_range'], tol=10, remove_text=True, extensions=['png']
)
def test_themis_available_multiday_time_range():
    time_range = ('2018-01-29T00:01:27', '2018-01-31T20:00')
    asilib.asi.plot_themis_available(time_range=time_range)
    return


import pathlib
import configparser

from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("asilib")  # Defined in pyproject.toml
except PackageNotFoundError:
    __version__ = "unknown"

# Load the configuration settings.
HERE = pathlib.Path(__file__).parent.resolve()
settings = configparser.ConfigParser()
settings.read(HERE / 'config.ini')

try:
    ASI_DATA_DIR = settings['Paths'].get('ASI_DATA_DIR', pathlib.Path.home() / 'asilib-data')
    ASI_DATA_DIR = pathlib.Path(ASI_DATA_DIR)
except KeyError:
    ASI_DATA_DIR = pathlib.Path.home() / 'asilib-data'

config = {'ASILIB_DIR': HERE, 'ASI_DATA_DIR': ASI_DATA_DIR, 'ACKNOWLEDGED_ASIS':[]}

# Imager implementation functions and classes.
from asilib.imager import Imager
from asilib.imagers import Imagers
from asilib.conjunction import Conjunction

__all__ = [
    'Imager',
    'Imagers',
    'Conjunction',
]  # So Sphinx shortens the name. See https://stackoverflow.com/a/31594545



"""
A collage to sell asilib to new users
"""
from datetime import datetime

import matplotlib.pyplot as plt
import matplotlib.gridspec
import numpy as np

import asilib.asi
import asilib.map


location_code = 'RANK'
time = datetime(2017, 9, 15, 2, 34, 0)
time_range = (datetime(2017, 9, 15, 2, 0, 0), datetime(2017, 9, 15, 4, 0, 0))
map_alt_km = 110
fontsize = 17

lon_bounds = (-102, -82)
lat_bounds = (56, 68)

# Create plots.
# ax is a 2x2 subplot grid with the fisheye and map images from THEMIS and REGO.
# bx and cx subplots are the corresponding keograms.
fig = plt.figure(figsize=(4.5, 7))
gs = matplotlib.gridspec.GridSpec(3, 2, fig)
ax = np.zeros((2, 2), dtype=object)
ax[0, 0] = plt.subplot(gs[0, 0])
ax[1, 0] = plt.subplot(gs[1, 0])
ax[0, 1] = asilib.map.create_map(
    lon_bounds=lon_bounds, lat_bounds=lat_bounds, fig_ax=(fig, gs[0, 1])
)
ax[1, 1] = asilib.map.create_map(
    lon_bounds=lon_bounds, lat_bounds=lat_bounds, fig_ax=(fig, gs[1, 1])
)
bx = plt.subplot(gs[2, :])
# cx = plt.subplot(gs[3,:])

themis1 = asilib.asi.themis(location_code, time=time)
themis1.plot_fisheye(ax=ax[0, 0], label=False)
themis1.plot_map(ax=ax[0, 1], asi_label=False)
rego1 = asilib.asi.rego(location_code, time=time)
rego1.plot_fisheye(ax=ax[1, 0], label=False)
rego1.plot_map(ax=ax[1, 1], asi_label=False)

themis2 = asilib.asi.themis(location_code, time_range=time_range)
themis2.plot_keogram(ax=bx, title=False)
# rego2 = asilib.asi.rego(location_code, time_range=time_range)
# rego2.plot_keogram(ax=cx, title=False)

ax[0, 0].axis('off')
ax[1, 0].axis('off')
bx.axis('off')

plt.suptitle(f'asilib', fontsize=20)
plt.subplots_adjust(top=0.941, bottom=0.01, left=0.023, right=0.977, hspace=0.1, wspace=0.11)
plt.show()



"""
This example generates a keogram of a STEVE aurora that moved equatorward. This event was studied
in:

Gallardo-Lacourt, B., Nishimura, Y., Donovan, E., Gillies, D. M., Perry, G. W., Archer, W. E., et al. (2018). 
A statistical analysis of STEVE. Journal of Geophysical Research: Space Physics, 123, 9893-9905. 
https://doi.org/10.1029/2018JA025368
"""

import matplotlib.pyplot as plt

import asilib.asi

location_code = 'LUCK'
time_range = ['2017-09-27T07', '2017-09-27T09']
map_alt_km = 230

fig, ax = plt.subplots(figsize=(8, 6))
asi = asilib.asi.rego(location_code, time_range=time_range, alt=map_alt_km)
ax, p = asi.plot_keogram(ax=ax, color_bounds=(300, 800), aacgm=True)
plt.colorbar(p, label='Intensity')
ax.set_xlabel('UTC')
ax.set_ylabel(f'Magnetic Latitude [deg]\nEmission altitude={map_alt_km} km')
plt.tight_layout()
plt.show()



"""
Maps an image of STEVE (the thin band). Reproduced from 
http://themis.igpp.ucla.edu/nuggets/nuggets_2018/Gallardo-Lacourt/fig2.jpg
"""

from datetime import datetime

import matplotlib.pyplot as plt

import asilib.asi
import asilib.map

ax = asilib.map.create_map(lon_bounds=(-127, -100), lat_bounds=(45, 65))

asi = asilib.asi.themis('ATHA', time=datetime(2010, 4, 5, 6, 7, 0), alt=110)
asi.plot_map(ax=ax)
plt.tight_layout()
plt.show()



"""
Tests the asilib's Downloader class.
"""
import os
import tempfile
from datetime import datetime

import pytest
import requests

import asilib
from asilib.download import Downloader


def test_bad_url():
    """
    Checks that Downloader() raises an error for a non-existant URL.
    """
    d = Downloader('https://data.phys.ucalgary.ca/DOES_NOT_EXIST_9876')
    with pytest.raises(requests.exceptions.HTTPError):
        d.ls()
    return


def test_ls_and_download():
    """
    Tests the list files (ls) method, .url and .name attributes,
    and the download method when navigating down the rather complex
    https://data.phys.ucalgary.ca/sort_by_project/THEMIS/asi/stream0/
    folder structure.
    """
    date = datetime(2014, 5, 5, 5, 10)
    location = 'gill'

    base_url = (
        f'https://data.phys.ucalgary.ca/sort_by_project/THEMIS/asi/stream0/'
        f'{date.year}/{date.month:02}/{date.day:02}/'
    )
    d = Downloader(base_url, headers={'User-Agent':'asilib'})

    # Check that Downloader.ls finds the unique online subdirectory
    matched_downloaders = d.ls(f'{location.lower()}_themis*')
    assert len(matched_downloaders) == 1
    assert matched_downloaders[0].url == (
        'https://data.phys.ucalgary.ca/'
        'sort_by_project/THEMIS/asi/stream0/2014/05/05/gill_themis19/'
    )

    # Navigate further down the subdirectory structure and find the file.
    d2 = Downloader(matched_downloaders[0].url + f'ut{date.hour:02}/', headers={'User-Agent':'asilib'})
    file_search_str = f'{date.strftime("%Y%m%d_%H%M")}_{location.lower()}*.pgm.gz'
    matched_downloaders2 = d2.ls(file_search_str)

    assert len(matched_downloaders2) == 1
    assert matched_downloaders2[0].url == (
        'https://data.phys.ucalgary.ca/'
        'sort_by_project/THEMIS/asi/stream0/2014/05/05/gill_themis19/ut05/'
        '20140505_0510_gill_themis19_full.pgm.gz'
    )
    assert matched_downloaders2[0].name == '20140505_0510_gill_themis19_full.pgm.gz'

    with tempfile.TemporaryDirectory() as tmp:
        download_path = matched_downloaders2[0].download(tmp, redownload=True)
        assert download_path.name == '20140505_0510_gill_themis19_full.pgm.gz'
        assert os.path.getsize(download_path) == 1957270
    return


def test_download_file_no_subdirectories():
    """
    Test Download.download() for the case where no subdirectories were provided (the filename)
    is in the base_url.
    """
    d = Downloader(
        'https://data.phys.ucalgary.ca/sort_by_project/THEMIS/asi/stream0/'
        '2014/05/05/gill_themis19/ut08/20140505_0807_gill_themis19_full.pgm.gz',
        headers={'User-Agent':'asilib'}
    )
    with tempfile.TemporaryDirectory() as tmp:
        download_path = d.download(tmp, redownload=True)
        assert download_path.name == '20140505_0807_gill_themis19_full.pgm.gz'
        assert os.path.getsize(download_path) == 1980541
    return



from datetime import datetime, timedelta
import dateutil.parser

import pytest

import asilib.utils as utils

str_time = '2010-01-01T10:29:32.000000'
obj_time = datetime(2010, 1, 1, 10, 29, 32, 0)
valid_times = [(str_time, obj_time), (obj_time, obj_time)]


@pytest.mark.parametrize("test_input,expected", valid_times)
def test_validate_time(test_input, expected):
    assert utils.validate_time(test_input) == expected


invalid_times = [1, 1.0]


@pytest.mark.parametrize("test_input", invalid_times)
def test_invalid_validate_time(test_input):
    with pytest.raises(ValueError):
        utils.validate_time(test_input)


def test_invalid_validate_time_year():
    with pytest.raises(ValueError):
        utils.validate_time(datetime(1900, 1, 1, 10, 29, 32, 0))
    with pytest.raises(ValueError):
        future_year = datetime.now().year + 1
        utils.validate_time(datetime.now().replace(year=future_year))


valid_time_ranges = [
    ([str_time, str_time], [obj_time, obj_time]),
    ([str_time, obj_time], [obj_time, obj_time]),
    ([obj_time, obj_time], [obj_time, obj_time]),
]


@pytest.mark.parametrize("test_input,expected", valid_time_ranges)
def test_validate_time_range(test_input, expected):
    assert utils.validate_time_range(test_input) == expected


invalid_time_ranges = [[], [obj_time], [obj_time, obj_time, obj_time]]


@pytest.mark.parametrize("test_input", invalid_time_ranges)
def test_validate_time_range_invalid(test_input):
    with pytest.raises(AssertionError):
        utils.validate_time_range(test_input)


def test_get_filename_times_partial_day():
    time_range = (datetime(2010, 1, 1), datetime(2010, 1, 5, 12))
    file_times = [
        datetime(2010, 1, 1),
        datetime(2010, 1, 2),
        datetime(2010, 1, 3),
        datetime(2010, 1, 4),
        datetime(2010, 1, 5),
    ]
    dt = 'days'
    assert utils.get_filename_times(time_range, dt=dt) == file_times


def test_get_filename_times_end_day():
    time_range = (datetime(2010, 1, 1), datetime(2010, 1, 5))
    file_times = [
        datetime(2010, 1, 1),
        datetime(2010, 1, 2),
        datetime(2010, 1, 3),
        datetime(2010, 1, 4),
    ]
    dt = 'days'
    assert utils.get_filename_times(time_range, dt=dt) == file_times


def test_get_filename_times_partial_hour():
    time_range = (datetime(2010, 1, 1, 10, 29, 32, 0), datetime(2010, 1, 1, 12, 29, 32, 0))
    file_times = [
        datetime(2010, 1, 1, 10, 0, 0, 0),
        datetime(2010, 1, 1, 11, 0, 0, 0),
        datetime(2010, 1, 1, 12, 0, 0, 0),
    ]
    dt = 'hours'
    assert utils.get_filename_times(time_range, dt=dt) == file_times


def test_get_filename_times_end_hour():
    time_range = (datetime(2010, 1, 1, 10, 0, 0, 0), datetime(2010, 1, 1, 12, 0, 0, 0))
    file_times = [datetime(2010, 1, 1, 10, 0, 0, 0), datetime(2010, 1, 1, 11, 0, 0, 0)]
    dt = 'hours'
    assert utils.get_filename_times(time_range, dt=dt) == file_times


def test_get_filename_times_partial_minute():
    time_range = (datetime(2010, 1, 1, 10, 29, 32, 0), datetime(2010, 1, 1, 10, 31, 32, 0))
    file_times = [
        datetime(2010, 1, 1, 10, 29, 0, 0),
        datetime(2010, 1, 1, 10, 30, 0, 0),
        datetime(2010, 1, 1, 10, 31, 0, 0),
    ]
    dt = 'minutes'
    assert utils.get_filename_times(time_range, dt=dt) == file_times


def test_get_filename_times_end_minute():
    time_range = (datetime(2010, 1, 1, 10, 29, 32, 0), datetime(2010, 1, 1, 10, 31, 0, 0))
    file_times = [
        datetime(2010, 1, 1, 10, 29, 0, 0),
        datetime(2010, 1, 1, 10, 30, 0, 0),
    ]
    dt = 'minutes'
    assert utils.get_filename_times(time_range, dt=dt) == file_times



"""
This is an example aurora due to a field line resonance studied in:

Gillies, D. M., Knudsen, D., Rankin, R., Milan, S., & Donovan, E. (2018). A statistical survey
of the 630.0-nm optical signature of periodic auroral arcs resulting from magnetospheric field
line resonances. Geophysical Research Letters, 45(10), 4648-4655.
"""
import matplotlib.pyplot as plt

import asilib.asi

location_code = 'GILL'
time_range = ['2015-02-02T10', '2015-02-02T11']

asi = asilib.asi.rego(location_code, time_range=time_range, alt=230)
ax, p = asi.plot_keogram(color_map='Greys_r')
plt.xlabel('Time')
plt.ylabel('Geographic Latitude [deg]')
plt.colorbar(p)
plt.tight_layout()
plt.show()



import asilib.asi

import matplotlib.testing.decorators
import matplotlib.pyplot as plt

from asilib.imager import Skymap_Cleaner

@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_geodetic_skymap'],
    tol=20,
    remove_text=True,
    extensions=['png'],
)
def test_geodetic_skymap():
    """
    Compare the (lat, lon) skymaps between the asilib and the official implementation.
    """
    time = '2020-01-01'
    ref_asi = asilib.asi.themis('GILL', time=time, load_images=False, alt=110)
    asi = asilib.asi.themis('GILL', time=time, load_images=False, custom_alt=True, alt=110)

    _ref_cleaner = Skymap_Cleaner(ref_asi.skymap['lon'], ref_asi.skymap['lat'], ref_asi.skymap['el'])
    _ref_lon, _ref_lat = _ref_cleaner.remove_nans()

    _custom_cleaner = Skymap_Cleaner(asi.skymap['lon'], asi.skymap['lat'], asi.skymap['el'])
    _custom_lon, _custom_lat = _custom_cleaner.remove_nans()

    _, ax = plt.subplots(2, sharex=True, sharey=True, figsize=(4, 8))
    ax[0].pcolormesh(_ref_lon, _ref_lat, _ref_lat)
    ax[1].pcolormesh(_custom_lon, _custom_lat, _custom_lat)
    return